/*
 *  generate.c
 *  dev
 *
 *  Created by laurent joyeux on 31/03/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */

#include <BaseLib/BaseLib.h>
#include <VideoLib/VideoLib.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/times.h>

typedef struct 
{
    PathComponents_t    *path;
    TimeCode_t          tc;
} MovieMD_t;


MovieMD_t** readMoviesFile(char *fileName)
{
    char        line[256];
    MovieMD_t   **movies;
    int         allocated;
    int         index;
    
    FILE *file = fopen(fileName, "r");
    
    if( NULL == file )
        return NULL;
    
    movies = NULL;
    allocated = index = 0;
    while( fgets(line, 256, file) != NULL )
    {
        const char *seps = " \t";
        char * fileName = strtok(line, seps);
        char * duration = strtok(NULL, seps);
        
        if( index==allocated )
        {
            allocated += 10;
            movies = realloc( movies, sizeof(MovieMD_t*)*allocated );
            int i;
            for(i=index; i<allocated; i++)
                movies[i] = NULL;
        }
        
        MovieMD_t *movie = (MovieMD_t*) bsMalloc( sizeof(MovieMD_t) );
        movies[index++] = movie;
        
        movie->path = bsPathCompInit( fileName );
        tcStringToTC( & movie->tc, duration);
    }
    
    fclose( file );
    
    return movies;
}

const MovieMD_t * getMovie(const MovieMD_t **movies, const char *fileName)
{
    PathComponents_t *path = bsPathCompInit( fileName );
    char * name = bsPathCompGetFileName( path );
    const MovieMD_t   *movie;
    int         index;
    
    movie = NULL;
    for(index=0; movies[index]!=NULL; index++)
    {
        if( !strcmp(name, bsPathCompGetFileName( movies[index]->path ) ) )
        {
            movie = movies[index];
            break;
        }
    }
    
    bsPathCompFree( path );
    
    return movie;
}

char **readFileList(const char *fileName)
{
    char    **list = NULL;
    int     allocated = 0;
    int     index = 0;
    FILE    *file;
    char    line[256];
    
    file  = fopen(fileName, "r");
    
    if( NULL == file )
        return NULL;

    while( fgets( line, 256, file) != NULL )
    {
        if( index==allocated )
        {
            allocated += 10;
            list = (char **) realloc( list, sizeof(char*)*allocated );
            int i;
            for(i=index; i<allocated; i++)
                list[i] = NULL;
        }
        
        list[index++] = bsStrDup( line );
    }
    
    fclose( file );
    
    return list;
}

static void freePP(void **ptr, void (*fct)(void *))
{
    int index;
    for(index=0; ptr[index]; index++)
        fct(ptr[index]);
    
    bsFree( ptr );
}

int main(int argc, char *argv[])
{
    char        **refGroup, **nonRefGroup;
    MovieMD_t   **movies;
    int         minLengthRef, maxLengthRef;
    int         minLengthNonRef, maxLengthNonRef;
    int         borderSrc, borderDst;
    int         numRequests;
    int         refGroupCount, nonRefGroupCount;

    if( argc!=9 )
    {
        fprintf(stderr, "%s refGroup nonRefGroup movieDuration minLengthRef,maxLengthRef minLengthNonRef,maxLengthNonRef beginAndEndRejectedMaterialAmount beginAndEndNonRefPresentMaterialAmount numRequests\n", argv[0]);
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");

        return 1;
    }
    
    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
    
    refGroup = readFileList(argv[1]);
    for(refGroupCount=0; refGroup[refGroupCount]!=NULL; refGroupCount++)  ;
//        printf("%s", group1[index]);

    nonRefGroup = readFileList(argv[2]);
    for(nonRefGroupCount=0; nonRefGroup[nonRefGroupCount]!=NULL; nonRefGroupCount++)  ;

    movies = readMoviesFile(argv[3]);
/*
    for(index=0; movies[index]!=NULL; index++)
        printf("%s\n", bsPathCompGetFileName(movies[index]->path));
*/  

    sscanf(argv[4], "%d,%d", &minLengthRef, &maxLengthRef);
    sscanf(argv[5], "%d,%d", &minLengthNonRef, &maxLengthNonRef);
    borderSrc = atoi( argv[6] );
    borderDst = atoi( argv[7] );
    numRequests = atoi( argv[8] );
    
    struct tms dummy_tms;
    srand48(times(&dummy_tms));
    int request;
    int count = 0;
    for(request=0; request<numRequests; request++)
    {
        int type;
        int randomRef = (int) (drand48()*refGroupCount);
        int randomNonRef = (int) (drand48()*nonRefGroupCount);
        int offsetRef, offsetNonRef, offset;
        int lengthRef, lengthNonRef; 
        
        if( count!=1000 ) 
            type = (int) (drand48()*3);


        const MovieMD_t *movieRef = getMovie( (const MovieMD_t **) movies, refGroup[randomRef] );
        const MovieMD_t *movieNonRef = getMovie( (const MovieMD_t **) movies, nonRefGroup[randomNonRef] );
        TimeCode_t tc;
        
        switch( type )
        {
            case 0:
                for(count=0;count<1000; count++)
                {
                    offsetRef = (int) (drand48()*(movieRef->tc.frameNumber-2*movieRef->tc.fps*borderSrc))+movieRef->tc.fps*borderSrc;
                    lengthRef = drand48()*(maxLengthRef-minLengthRef)+minLengthRef;
                    lengthRef *= movieRef->tc.fps;
                    
                    if( offsetRef+lengthRef < movieRef->tc.frameNumber-movieRef->tc.fps*borderSrc )
                        break;
                }
                if( 1000==count )
                {
                    request--;
                    continue;
                }

                tc.fps = movieRef->tc.fps;
                printf("%d %d null 0:0:0 0:0:0 %s ", request, type+1, bsPathCompGetPathFileName(movieRef->path));
                tcTCToFILE( tcFrameToTC( &tc, offsetRef),  stdout ); 
                printf(" ");
                tcTCToFILE( tcFrameToTC( &tc, offsetRef+lengthRef),  stdout ); 
                printf(" 0:0:0\n");

                break;
            case 1:
                for(count=0;count<1000; count++)
                {
                    lengthRef = drand48()*(maxLengthRef-minLengthRef)+minLengthRef;
                    lengthRef *= movieRef->tc.fps;
                    lengthNonRef = drand48()*(maxLengthNonRef-minLengthNonRef)+minLengthNonRef;
                    lengthNonRef *= movieNonRef->tc.fps;
                    if( lengthRef+2*movieRef->tc.fps*borderDst > lengthNonRef )
                        continue;
                        
                    offset = (int) (drand48()*(lengthNonRef - lengthRef-2*movieRef->tc.fps*borderDst)) + movieRef->tc.fps*borderDst;

                    offsetRef = (int) (drand48()*(movieRef->tc.frameNumber-2*movieRef->tc.fps*borderSrc))+movieRef->tc.fps*borderSrc;
                    offsetNonRef = (int) (drand48()*(movieNonRef->tc.frameNumber-2*movieNonRef->tc.fps*borderSrc))+movieNonRef->tc.fps*borderSrc;
                    
                    if( offsetRef+lengthRef > movieRef->tc.frameNumber-movieRef->tc.fps*borderSrc  )
                        continue;
                    if( offsetNonRef+lengthNonRef > movieNonRef->tc.frameNumber-movieNonRef->tc.fps*borderSrc  )
                        continue;
                        
                    break;
                }
                if( 1000==count )
                {
                    request--;
                    continue;
                }
                tc.fps = movieRef->tc.fps;
           
                printf("%d %d %s   ", request, type+1, bsPathCompGetPathFileName(movieNonRef->path));
                tcTCToFILE( tcFrameToTC( &tc, offsetNonRef),  stdout ); 
                printf(" ");
                tcTCToFILE( tcFrameToTC( &tc, offsetNonRef+lengthNonRef),  stdout ); 
                
                printf(" %s ", bsPathCompGetPathFileName(movieRef->path));
                tcTCToFILE( tcFrameToTC( &tc, offsetRef),  stdout ); 
                printf(" ");
                tcTCToFILE( tcFrameToTC( &tc, offsetRef+lengthRef),  stdout ); 
                printf(" ");
                tcTCToFILE( tcFrameToTC( &tc, offset),  stdout ); 
                printf("\n");
                
                break;
            case 2:
                for(count=0;count<1000; count++)
                {
                    offsetNonRef = (int) (drand48()*(movieNonRef->tc.frameNumber-2*movieNonRef->tc.fps*borderSrc))+movieNonRef->tc.fps*borderSrc;
                    lengthNonRef = drand48()*(maxLengthNonRef-minLengthNonRef)+minLengthNonRef;
                    lengthNonRef *= movieNonRef->tc.fps;
                    
                    if( offsetNonRef+lengthNonRef < movieNonRef->tc.frameNumber-movieNonRef->tc.fps*borderSrc )
                        break;
                }
                if( 1000==count )
                {
                    request--;
                    continue;
                }

                tc.fps = movieNonRef->tc.fps;
                printf("%d %d %s   ", request, type+1, bsPathCompGetPathFileName(movieNonRef->path));
                tcTCToFILE( tcFrameToTC( &tc, offsetNonRef),  stdout ); 
                printf(" ");
                tcTCToFILE( tcFrameToTC( &tc, offsetNonRef+lengthNonRef),  stdout ); 
                printf(" null 0:0:0 0:0:0 0:0:0\n");

                break;
        }
    }
    
    return 0;
}
