#! /usr/bin/perl

#
# (c) INRIA-IMEDIA 2008 (developed by L. Joyeux)
#

use strict;

my $arguments = "arguments: -d movieDuration -r requestMix -n nonRefGroup -i prefix_mix_movies -o prefix_transfomed_movies [-ot only_transform_X] [-t prefix_tmp_movies] [-m [min_request]:[max_request] ] [-b number_of_second_rejected_begin_end_nonRefGroup_movies] [-ns]\n";

my $tmpFolder = "";

my $durationFileName;
my $request1FileName;
my $nonRefGroupFileName;
my $outFolder;
my $dstFolder;
my $onlyTransfo;
my $tmpFolder;
my $border = 60*2;
my $nosound = 0;
my $requestInterval;

while(1)
{
    my $arg = shift;
    last if ! defined $arg;
    
    
    if( $arg eq '-d' )
    {
        $durationFileName = shift;
    }
    elsif ( $arg eq '-r' )
    {
        $request1FileName = shift;
    }
    elsif ( $arg eq '-n' )
    {
        $nonRefGroupFileName = shift;
    }
    elsif ( $arg eq '-i' )
    {
        $outFolder = shift;
    }
    elsif ( $arg eq '-o' )
    {
        $dstFolder = shift;
    }
    elsif ( $arg eq '-ot' )
    {
        $onlyTransfo = shift;
    }
    elsif ( $arg eq '-t' )
    {
        $tmpFolder = shift;
    }
    elsif ( $arg eq '-m' )
    {
        $requestInterval = shift;
    }
    elsif ( $arg eq '-b' )
    {
        $border = shift;
    }
    elsif ( $arg eq '-ns' )
    {
        $nosound = 1;
    }
    else
    {
        die "unrecognized option $arg\n$arguments";
    }
}

die "missing argument -d \n$arguments" if ( ! defined $durationFileName );
die "missing argument -r \n$arguments" if ( ! defined $request1FileName );
die "missing argument -n \n$arguments" if ( ! defined $nonRefGroupFileName );
die "missing argument -i \n$arguments" if ( ! defined $outFolder );
die "missing argument -o \n$arguments" if ( ! defined $dstFolder );

my @durations = read_duration($durationFileName);

my @requests = read_request1($request1FileName);

my @nonRefGroup = read_group($nonRefGroupFileName);

################################################### PARAMS ######################################################
my $width = 352; my $height = 288;

# transfo1
my %T1_angles;
$T1_angles{"kind"} = "samples";
$T1_angles{"values"} = [0, 20, 40];

# transfo2
my %T2_scale;
$T2_scale{"kind"} = "range";
$T2_scale{"values"} = "0.3:0.05:0.5";

my %T2_pos;
$T2_pos{"kind"} = "samples";
$T2_pos{"values"} = ["c", "lt", "lb", "rt", "rb"];

# transfo3
my %T3_images;
$T3_images{"kind"} = "samples";
$T3_images{"values"} = ["logo1.png", "logo2.png", "logo3.png", "logo4.png", "logo5.png", "logo6.png", "logo7.png", "logo8.png", "logo9.png", "logo10.png", "logo11.png", "logo12.png", "logo13.png", "logo14.png", "logo15.png"];

# transfo4
my %T4_scale;
$T4_scale{"kind"} = "range";
$T4_scale{"values"} = "3:1:7";

my %T4_bitrate;
$T4_bitrate{"kind"} = "samples";
$T4_bitrate{"values"} = [16000, 32000, 64000];

my %T4_codec;
$T4_codec{"kind"} = "samples";
$T4_codec{"values"} = ["mp4", "flv"];

# transfo5
my %T5_gamma;
$T5_gamma{"kind"} = "range";
$T5_gamma{"values"} = "0.3:0.1:1.8";
$T5_gamma{"reject"} = "0.9:1.1";

# transfo6
my %T_blur;
$T_blur{"kind"} = "range";
$T_blur{"values"} = "1:0.1:3";

my %T_drop;
$T_drop{"kind"} = "range";
$T_drop{"values"} = "0.05:0.05:0.2";

my %T_contrast;
$T_contrast{"kind"} = "range";
$T_contrast{"values"} = "-0.3:0.05:0.3";
$T_contrast{"reject"} = "-0.1:0.1";

my %T_ratio;
$T_ratio{"kind"} = "range";
$T_ratio{"values"} = "0.75:0.05:0.9";

my %T_noise_kind;
$T_noise_kind{"kind"} = "samples";
$T_noise_kind{"values"} = ["white"]; #, "gaussian"];

my %T_noise_amount;
$T_noise_amount{"kind"} = "range";
$T_noise_amount{"values"} = "0.1:0.05:0.2";

# transfo8
my %T8_crop;
$T8_crop{"kind"} = "range";
$T8_crop{"values"} = "5:5:20";

my %T8_shift_X;
$T8_shift_X{"kind"} = "range";
$T8_shift_X{"values"} = "-50:5:50";
$T8_shift_X{"reject"} = "-10:10";

my %T8_shift_Y;
$T8_shift_Y{"kind"} = "range";
$T8_shift_Y{"values"} = "-40:5:40";
$T8_shift_Y{"reject"} = "-10:10";

my %T8_text;
$T8_text{"kind"} = "samples";
$T8_text{"values"} = [];
push @{$T8_text{"values"}}, "text=\"copy detection\" font=./arial.ttf size=30 pos=center offset=" . $width/2 . ",200 fg=1,1,0,0 bg=0,0,0,0\n";
push @{$T8_text{"values"}}, "text=\" this is a caption line \" font=./arial.ttf size=30 pos=center offset=" . $width/2 . ",200 fg=1,1,1,1 bg=0,0,0,0\n";
push @{$T8_text{"values"}}, "text=\"inverted text line\" font=./arial.ttf size=30 pos=center offset=" . $width/2 . ",200 fg=0,0,0,0 bg=1,0,1,0\n";
push @{$T8_text{"values"}}, "text=\"BIG TEXT\" font=./arial.ttf size=60 pos=center offset=" . $width/2 . ",200 fg=1,0,0,0 bg=0,0,0,0\n";

#################################################################################################################
my %line = {};
my $counter=0;

my $transfoMin = 1;
my $transfoMax = 10;

if( defined $onlyTransfo )
{
    $transfoMin = $transfoMax = $onlyTransfo
}

my $minRequest = 0;
my $maxRequest = -1;

if( defined $requestInterval )
{
    if( $requestInterval =~ /:/ )
    {
        ($minRequest, $maxRequest) = split(/:/, $requestInterval);
        $minRequest= 0 if ( ""==$minRequest );
        $maxRequest= -1 if ( ""==$maxRequest );
    }
    else
    {
        $minRequest = $maxRequest = $requestInterval;
    }
}

for my $j (@requests)
{
    if ( $minRequest>$counter )
    {
        $counter++;
        next;
    }
    
    %line = %{$j};
    
    for(my $i=$transfoMin; $i<=$transfoMax; $i++)
    {
        
        my %params;
        
        $params{"input"} = $outFolder . $line{"id"} . ".mpg";;
        $params{"output"} = $dstFolder . $line{"id"} . "-" . $i . ".mpg";
        $params{"transfos"} = "";
        $params{"fileOpts"} = "";
        $params{"tmpIndex"} = 0;
        $params{"transformId"} = $i;
        $params{"fps"} = getFPS(\%line);
        $params{"size"} = [ split(/x/, "352x288") ];
       
        print "# Transfo $i - "  . $params{"input"} ."\n";
        
        $params{"transfos"} .= eval 'transfo'. $i . '(\%params)' ;
        
        flushTransfos(\%params);
        print "bitrate video=1000000\n";
        print "nosound\n" if $nosound;
        print "\n";

    }
        
    last if( (++$counter >= $maxRequest) && ( $maxRequest!=-1 ) );
}

sub flushTransfos
{
    my $params = shift;

    print "file input=" . $params->{"input"} . " output=" . $params->{"output"} . $params->{"fileOpts"} ."\n";
    print $params->{"transfos"};
    
    $params->{"fileOpts"} = "";
    $params->{"transfos"} = "";
}

sub generateTmp
{
    my $params = shift;

    return $tmpFolder . "tmp" . $params->{"transformId"} . "-" . ($params->{"tmpIndex"}++);
}

sub getFPS
{
    my $line  = shift;
    my @entries = ("nonRefTCIn", "nonRefTCOut", "refTCIn", "refTCOut", "refTCOffset");
    my $fps = 0;
    
    for my $j (@entries)
    {
        my $f = $line->{$j}{"fps"};
        if($f!=0)
        {
            die "line request with different fps" if $fps!=0 && $fps!=$f;
            $fps = $f;
        }
    }
        
    return $fps;
}

#########

sub transfo1
{
    my $a = random(\%T1_angles);
    return "angle=$a\n";
}

#########

sub transfo2_sub
{
    my $params = shift;

    my $scale = random(\%T2_scale);
    my $pos = random(\%T2_pos);
    my $x = 0;
    my $y = 0;
        
    if( $pos eq "c" )
    {
        $x = $width/2-$width/2*$scale;
        $y = $height/2-$height/2*$scale;
    }
    elsif( $pos eq "lt")
    {
        $x = 0;
        $y = 0;
    }
    elsif( $pos eq "lb")
    {
        $x = 0;
        $y = $height-$height*$scale;
    }
    elsif( $pos eq "rt")
    {
        $x = $width-$width*$scale;
        $y = 0;
    }
    elsif( $pos eq "rb")
    {
        $x = $width-$width*$scale;
        $y = $height-$height*$scale;
    }
    
    my $duration;
    my $path;
    
    if( 1==$line{"type"} )
    {
        $duration = $line{"refTCOut"}{"n"}-$line{"refTCIn"}{"n"};
    }
    else
    {
        $duration = $line{"nonRefTCOut"}{"n"}-$line{"nonRefTCIn"}{"n"};
        $path = $line{"nonRefName"};
    }
    
    my @els;
    my $file;
    if( defined($path) )
    {
        @els = split(/\//, $path);
        $file = $els[$#els];
    }
    
    my $nonRefDuration = -1;
	my $nonRefFps = 0;
    while()
    {
        my $r = int(rand()*$#nonRefGroup);
        my $path1 = $nonRefGroup[$r];
        $path1 =~ s/\n//;
    
        @els = split(/\//, $path1);
        my $file1 = $els[$#els];
        
        if( !defined($file) || ($file ne $file1) )
        {
            $path = $path1;
            
            my $d;
            for $d (@durations)
            {
                my %d1 = %{$d};
                @els = split(/\//, $d1{"name"});
                my $file2 = $els[$#els];

                if( $file2 eq $file1 )
                {
                    $nonRefDuration = $d1{"duration"}{"n"};
                    $nonRefFps = $d1{"duration"}{"fps"};
                    last;
                }
            }
        }

        next if $nonRefFps!=$params->{"fps"};
        
		if($nonRefDuration!=-1)
        {
            last if ($nonRefDuration-$duration-$border*$params->{"fps"}*2 > 0);
        }        
    }
    my $offset = int(rand()*($nonRefDuration-$duration-$border*$params->{"fps"}*2) + $border*$params->{"fps"});

    my %r;
    $r{"x"} = $x;
    $r{"y"} = $y;
    $r{"offset"} = $offset;
    $r{"duration"} = $duration;
    $r{"file"} = $path;
    $r{"scale"} = $scale;
    
    return %r;
}


sub transfo2
{
    my $params = shift;
   
    my %r = transfo2_sub($params);
    
    my $tmp = $params->{"input"};
    
    my $tmpFile = generateTmp($params) . ".mp4";
    
    if( $params->{"transfos"} ne "" )
    {
        my $output = $params->{"output"};
        $params->{"output"} = $tmpFile;
        flushTransfos($params);
        $params->{"output"} = $output;
        
        print "\n";
        
        $tmp = $tmpFile;
        $params->{"transfos"} = "";
    }

    $params->{"input"} = $r{"file"};
    $params->{"fileOpts"} = " tcIn=" . frameNumber2Tc($r{"offset"}, $params->{"fps"}) . " tcOut=". frameNumber2Tc($r{"offset"}+$r{"duration"}, , $params->{"fps"});

    my $x = int($r{"x"});
    my $y = int($r{"y"});
    my $scale = $r{"scale"};
    
    return "pictInPict movie=$tmp  scale=$scale,$scale offset=$x,$y\n";
}

sub transfo2V2
{
    my ($params) = @_;
    
    my %r = transfo2_sub($params);
    
    my $x = int($r{"x"});
    my $y = int($r{"y"});
    my $scale = $r{"scale"};
    
    return "pictInPict movie=" . $r{"file"} . " scale=$scale,$scale offset=$x,$y tcOffset=" . frameNumber2Tc($r{"offset"}, $params->{"fps"}) . "\n";
}

#########

sub transfo3
{
    my $image = random(\%T3_images);
    return "pictInPict image=$image\n";
    
}

#########

sub transfo4
{
    my $params = shift;
    
    my $resample = random(\%T4_scale);
    my $bitrate = random(\%T4_bitrate);
    my $codec = random(\%T4_codec);
    my $scale = -1;
    my $transform;
    
    my $tmpName = generateTmp($params) . ".$codec";
    
    my $output = $params->{"output"};
    $params->{"output"} = $tmpName;
    flushTransfos($params);
    $params->{"output"} = $output;
    
    if( $resample>4 )
    {
        $scale = $resample/4;
        $resample = 4;
    }
    
    print "scale=" . (1/$scale) . "," . (1/$scale) . "\n" if ($scale>0); 
    print "resample=$resample\n";
    print "bitrate video=$bitrate\n";
    print "\n";
    
    $params->{"input"} = $tmpName;
    
    $transform = "";
    $transform .="resample size=$width,$height\n";
    $transform .= "scale=$scale,$scale\n" if ($scale>0); 
    
    return $transform;
}

#########

sub transfo5
{
    my $g;
    for(;;)
    {
        $g = random(\%T5_gamma);
        last if $g!=1.0;
    }
    
    return "gamma=$g,$g,$g\n";
}

#########

sub T6_blur
{
    my $sigma = random(\%T_blur);
    return "blur=$sigma\n";
}

sub T6_drop
{
    my $frequency = random(\%T_drop);
    return "drop=$frequency\n";
}

sub T6_contrast
{
    my $c = random(\%T_contrast);
    return "level constrast=$c\n";
}

sub T6_ratio
{
    my $r = random(\%T_ratio);
    return "scale=1,$r\n";


}

sub T6_noise
{
    my $params = shift;

    my $k = random(\%T_noise_kind);
    my $v = random(\%T_noise_amount);
    
    my $tmpName = generateTmp($params) . ".mp4";
    
    my $output = $params->{"output"};
    $params->{"output"} = $tmpName;
    
    $params->{"transfos"} .= "noise kind=$k,$k,$k stddev=$v,$v,$v\n";
    $params->{"transfos"} .= "bitrate video=2000000\n";
    
    flushTransfos($params);
    $params->{"output"} = $output;
    print "\n";

    $params->{"input"} = $tmpName;

    return "";
}

sub transfo6
{
    my $params = shift;
    
    my @t = randomV2(3, 7);
    my @subs = ('T6_blur', 'transfo5', 'T6_drop', 'T6_contrast', 'T6_ratio', 'T6_noise', 'transfo4');
    
    my $myTransfos = "";
    
    for my $i (@t)
    {
        my $v = eval $subs[$i] . '($params)';
        $params->{"transfos"} .= $v; 
    }
    
    return "";
}

#########

sub transfo7
{
    my $params = shift;
    
    my @t = randomV2(5, 7); #3
    my @subs = ('T6_blur', 'transfo5', 'T6_drop', 'T6_contrast', 'T6_ratio', 'T6_noise', 'transfo4');
    
    my $myTransfos = "";
    
    for my $i (@t)
    {
        my $v = eval $subs[$i] . '($params)';
        $params->{"transfos"} .= $v; 
    }
    
    return "";
}

#########

sub T8_crop
{
    my $cx = random(\%T8_crop);
    my $cy = random(\%T8_crop);
    
    my $w = ($width-$cx*2);
    my $h = ($height-$cy*2);
    
    return "crop offset=$cx,$cy size=$w,$h\n";
}

sub T8_shift
{
    my $sx = random(\%T8_shift_X);
    my $sy = random(\%T8_shift_Y);

    return "shift=$sx,$sy\n";    
}

sub T8_flip
{
    return "flip=1\n";    
}

sub T8_caption
{
    my $params = shift;
    
    my $t = random(\%T8_text);
    my $width = $params->{"width"};
    my $height = $params->{"height"};
    
    return eval $t ;
}

sub T8_caption_or_logo
{
    my $params = shift;
    my $r = rand();
    my @s1 = @{$T3_images{"values"}};
    my @s2 = @{$T8_text{"values"}};
    my $count = $#s1+$#s2+2;
    my $r = rand()*$count;
    
    if( $r>($#s1+1) )
    {
        return T8_caption($params);
    }
    else
    {
        return transfo3($params);
    }
}

sub transfo8
{
    my $params = shift;
    my @t = randomV2(3, 7);
    my @subs = ('T8_crop', 'T8_shift', 'T6_contrast', 'T8_caption', 'T8_flip', 'transfo2V2', 'transfo3' );
    
    for my $i (@t)
    {
        my $v = eval $subs[$i] . '($params)';
        $params->{"transfos"} .= $v; 
    }
    
    return "";

}

#########

sub transfo9
{
    my $params = shift;
    my @t = randomV2(5, 7);
    my @subs = ('T8_crop', 'T8_shift', 'T6_contrast', 'T8_caption', 'T8_flip', 'transfo2V2', 'transfo3' );
    
    for my $i (@t)
    {
        my $v = eval $subs[$i] . '($params)';
        $params->{"transfos"} .= $v; 
    }
    
    return "";

}

#########

sub transfo10
{
    my $params = shift;
    my @t;
    my @subs = ('transfo2', 'transfo2V2', 'transfo3', 'transfo4', 'transfo5', 'T6_blur', 'T6_drop', 'T6_ratio', 'T6_noise', 'T8_crop', 'T8_shift', 'T6_contrast', 'T8_caption', 'T8_flip');
    
    while(1)
    {
        @t = randomV2(5, 14);
        my @ordered = sort{ $a <=> $b } @t;
        last if( !(0==$ordered[0] && 1==$ordered[1]) );
    }
    
    my $i;
    for $i (@t)
    {
        my $v = eval $subs[$i] . '($params)';
        $params->{"transfos"} .= $v; 
    }
    
    return "";

}

#################################################################################################################
sub random
{
    my $v = shift;
    my $kind = $v->{"kind"};

    
    if( $kind eq "samples" )
    {
        my @samples = @{$v->{"values"}};
        my $r = rand($#samples+1);
        
        return  $samples[$r];
        
    }
    else
    {
        my ($a, $b, $c) = split(/:/, $v->{"values"});
        my ($r1, $r2) = split(/:/, $v->{"reject"});
        
        if( !defined($c) )
        {
            while(1)
            {
                my $r = rand()*($b-$a)+$a;
                return $r if ( !defined($r1) || !defined($r2) );
                return $r if ($r<$r1 || $r>$r2);
            }
        }
        else
        {
            while(1)
            {
                my $r = int(rand()*($c-$a)/$b+1)*$b+$a;
                return $r if ( !defined($r1) || !defined($r2) );
                return $r if ($r<$r1 || $r>$r2);
            }
        }
    }
}

sub randomV2
{
    my $m = shift;
    my $n = shift;
    my @a = ();
    
    for(;;)
    {
        my $r = int(rand()*$n);
        my $found = 0;
        my $i;
        foreach $i (@a)
        {
            if( $i == $r )
            {
                $found = 1;
                last;
            }
        }

        if( $found==0 )
        {
            push @a, $r;
            $m--;
            
            if( $m==0 )
            {
#                @els = sort{ $a <=> $b } @a;
#                return @els;
                return @a;
            }
            
        }
    }
}

sub read_duration
{
	my $filename = shift;
	open FILE, $filename;
	
	my @list;
	while(<FILE>)
	{
        my $l = $_;
        my @els = split(/ |\t/, $l);

        my %line = {};
        $line{"name"} = @els[0];
        $line{"duration"} = {tc_decode($els[1])};
        $line{"size"} = $els[2];

		push @list, {%line};
	}
	close FILE;

	return @list;
}

sub read_request1
{
	my $filename = shift;
	open FILE, $filename;
	
	my @list;
	while(<FILE>)
	{
        my $l = $_;
        my @els = split(/\s+/, $l);
        
        my %line = {};
        $line{"id"} = $els[0];
        $line{"type"} = $els[1];
        $line{"nonRefName"} = $els[2];
        $line{"nonRefTCIn"} = {tc_decode($els[3])};
        $line{"nonRefTCOut"} = {tc_decode($els[4])};
        $line{"refName"} = $els[5];
        $line{"refTCIn"} = {tc_decode($els[6])};
        $line{"refTCOut"} = {tc_decode($els[7])};
        $line{"refTCOffset"} = {tc_decode($els[8])};

		push @list, {%line};
	}
	close FILE;

	return @list;
}

sub read_group
{
	my $filename = shift;
	open FILE, $filename;
	
	my @list;
	while(<FILE>)
	{   
		my $l = $_;
		$l =~ s/\n$//;
		$l =~ s/\r$//;
		
		my @els = split(/\s+/, $l);
		
		push @list, @els;
	}
	close FILE;
	
	return @list;
}

sub tc_decode
{
    my $tc = shift;
    my @els = split(/:|\//, $tc);
	my $h = int($els[0]);
	my $m = int($els[1]);
	my $s = int($els[2]);
	my $f = int($els[3]);
	my $fps = 1.0*$els[4];
    my $n = (($h*60+$m)*60+$s)*$fps+$f;
    
	my %tc_hash = {};
	
	$tc_hash{"h"}   = $h;
	$tc_hash{"m"}   = $m;
	$tc_hash{"s"}   = $s;
	$tc_hash{"f"}   = $f;
	$tc_hash{"fps"} = $fps;
	$tc_hash{"n"}   = $n;
	
	return %tc_hash;
}

sub frameNumber2Tc
{
    my $n = shift;
    my $fps = shift;
    
    my $s = int($n/$fps);
    $n -= $s*$fps;
    my $m = int($s/60);
    $s -= $m*60;
    my $h = int($m/60);
    $m -= $h*60;
    
    return $h . ":" . $m . ":" . $s . ":" . $n . "/$fps";
}
