/*
 *  mergeMovies.c
 *  dev
 *
 *  Created by laurent joyeux on 01/04/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */

#include <BaseLib/BaseLib.h>
#include <VideoLib/VideoLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main( int argc, char *argv[] )
{
    FILE    *file;
    
    if( argc<1 )
    {
        fprintf( stderr, "%s movies\n", argv[0] );
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
        
        return 1;
    }

    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");

    int frame;
    int loop;
    Image_t *pimg;
    Image_t *pimg_red;
    SoundBuffer_t *psb = NULL;
    ImageSequenceWriter_t *pisw = NULL;
    
    pimg = NULL;
    pimg_red = NULL;
    
    TimeCode_t tc;
    tc.fps = 25;
    char tcString[128];

    file = fopen("mergeTC.txt", "w");
    frame = 0;
    
    for(loop=1; loop<argc; loop++)
    {
        ImageSequenceReader_t *pisr = isrOpen(argv[loop]);
        
        if( NULL==pisr )
        {
            fprintf(stderr, "Cannot open %s\n", argv[loop]);
            abort();
        }
        printf("%s %d/%d\n", argv[loop], loop, argc);
        tcFrameToTC(&tc, frame);
        tcTCToString(&tc, tcString);
        
        fprintf(file, "%s %s", argv[loop], tcString);
        
        while( !isrGetFrame( pisr, &pimg, NULL ) )
        {
            if( NULL==pisw )
            {
                char    buffer[256];
                BSDictionary_t *props;
                
                props = bsDictNew();

                sprintf( buffer, "%d,%d", imgGetWidth( pimg ), imgGetHeight( pimg ) );
                
                bsDictSetValueForKey( props, bsStrDup(buffer), "video_size",    BSDICTFREE_VALUE );
                //    bsDictSetValueForKey( props, "mpeg2video",  "video_codec",      BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "1000000",     "video_bitrate",    BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "25",          "video_fps",        BSDICTFREE_NONE );

                if( psb!= NULL )
                {
                    sprintf( buffer, "%d", psb->properties.sampleRate );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_rate", BSDICTFREE_VALUE );
                    
                    sprintf( buffer, "%d", psb->properties.numChannels );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_num_channels", BSDICTFREE_VALUE );
                    
                    bsDictSetValueForKey( props, "128000", "audio_bitrate", BSDICTFREE_NONE );
                    
                }
                
                sprintf( buffer, "merged.mpg" );
                pisw = iswOpen( buffer, props );
                bsDictFree( props );
            }
            
            iswPutFrame( pisw, pimg, NULL );
            frame++;
        }
        isrClose( pisr );

        tcFrameToTC(&tc, frame);
        tcTCToString(&tc, tcString);
        
        fprintf(file, " %s\n", tcString);
        
        if( NULL==pimg_red )
        {
            imgCopy(pimg, &pimg_red);
            uint8_t *pu8 = imgGetTopLeftPixel( pimg_red );
            int width = imgGetWidth( pimg_red );
            int height = imgGetHeight( pimg_red );
            int i, j;
            int bytesPerRow = imgGetBytesPerRow( pimg_red );
            
            for(j=0; j<height; j++, pu8+=bytesPerRow)
            {
                for(i=0; i<width; i++)
                {
                    pu8[3*i+0] = 255;
                    pu8[3*i+1] = 0;
                    pu8[3*i+2] = 0;
                }
            }
        }
        
        int counter;
        for(counter=0; counter<25*5; counter++)
        {
            iswPutFrame( pisw, pimg_red, NULL );
            frame++;
        }
        
    }
    
    imgFree( pimg );
    imgFree( pimg_red );
    
    iswClose( pisw );
    fclose( file );
    
    return 0;
}
