/*
 *  copy.c
 *  dev
 *
 *  Created by laurent joyeux on 01/04/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */

#include <BaseLib/BaseLib.h>
#include <VideoLib/VideoLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main( int argc, char *argv[] )
{
    FILE    *file;
    char    line[512];
    char    *parts[9];
    Image_t *pimg_Out;
    
    if( argc<3 )
    {
        fprintf( stderr, "%s movie_file dst_folder [widthxheight]\n", argv[0] );
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
        
        return 1;
    }
    
    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");

    pimg_Out = NULL;
    
    file = fopen( argv[1], "r" );
    while( fgets( line, 512, file ) != NULL )
    {
        char    *src = line;
        int     i;
        
        for( i=0; i<9; i++, src=NULL )
            parts[i] = strtok( src, " \t\n" );
        
        for( i=0; i<9; i++ )
            printf( "%s ", parts[i] );
        
        printf( "\n" );
        
        int type;
        
        char *fileNameNonRef;
        char *fileNameRef;
        TimeCode_t tcInNonRef, tcOutNonRef, tcInRef, tcOutRef, tcOffset;
    
        type = atoi( parts[1] );
        
        fileNameNonRef = parts[2];
        tcStringToTC( &tcInNonRef, parts[3] );
        tcStringToTC( &tcOutNonRef, parts[4] );
        
        fileNameRef = parts[5];
        tcStringToTC( &tcInRef, parts[6]) ;
        tcStringToTC( &tcOutRef, parts[7] );
        tcStringToTC( &tcOffset, parts[8] );

        ImageSequenceReader_t *pisr_NonRef, *pisr_Ref;
        ImageSequenceWriter_t *pisw;

	pisr_NonRef = NULL;
	pisr_Ref = NULL;

        SoundBuffer_t *psb_Ref, *psb_NonRef, *psb;
        psb_Ref = psb_NonRef = NULL;

        pisr_Ref = pisr_NonRef = NULL;
        if( 1==type || 2==type )
        {
            pisr_Ref = isrOpen( fileNameRef );
                
            isrSeekToFrame( pisr_Ref, (double)tcInRef.frameNumber / tcInRef.fps );
        }

        if( 2==type || 3==type )
        {
            pisr_NonRef = isrOpen( fileNameNonRef );
            isrSeekToFrame( pisr_NonRef, (double)tcInNonRef.frameNumber / tcInNonRef.fps );
        }
        
        Image_t *pimg_Ref, *pimg_NonRef, *pimg;
        pimg_Ref = pimg_NonRef = NULL;
        pisw = NULL;
        
        
        
        int     frame;
        int     stop = 0;
        for( frame=0; !stop; frame++ )
        {
            printf( "% 5d   \r", frame ); fflush( stdout );

            // type 2 or type 3 : read unconditionally non ref 
            if( 2==type || 3==type )
            {
                if( isrGetFrame( pisr_NonRef, &pimg_NonRef, &psb_NonRef ) < 0 )
                {
                    fprintf(stderr, "\n%s:%d abort\n", __FILE__, __LINE__ );
                    abort();
                }
                
                pimg = pimg_NonRef;
                psb  = psb_NonRef;
                
                stop = frame>=(tcOutNonRef.frameNumber - tcInNonRef.frameNumber);
                
                if( 2==type && frame>=tcOffset.frameNumber && frame<tcOffset.frameNumber+tcOutRef.frameNumber-tcInRef.frameNumber )
                {
                    if( isrGetFrame( pisr_Ref, &pimg_Ref, &psb_Ref ) < 0 )
                    {
                        fprintf(stderr, "\n%s:%d abort\n", __FILE__, __LINE__ );
                        abort();
                    }
                    pimg = pimg_Ref;
                    psb  = psb_Ref;
                }
            }
            else
            {
                if( isrGetFrame( pisr_Ref, &pimg_Ref, &psb_Ref ) < 0 )
                {
                    fprintf(stderr, "\n%s:%d abort\n", __FILE__, __LINE__ );
                    abort();
                }
                
                pimg = pimg_Ref;
                psb  = psb_Ref;
                stop = frame>=(tcOutRef.frameNumber - tcInRef.frameNumber);
            }
            
            if( (4 == argc) && (NULL == pimg_Out) )
            {
                int width, height;
                if( 2==sscanf( argv[3], "%dx%d", &width, &height ) )
                    pimg_Out = imgInit( width, height, imgGetFormat( pimg ), imgGetType( pimg ) );
            }
            
            if( pimg_Out != NULL )
            {
                imgResize( pimg, pimg_Out, IMGRESIZE_BILINEAR );
                pimg = pimg_Out;
            }
            
            if( NULL==pisw )
            {
                char    buffer[256];
                BSDictionary_t *props;
                
                props = bsDictNew();

                sprintf( buffer, "%d,%d", imgGetWidth( pimg ), imgGetHeight( pimg ) );
                
                bsDictSetValueForKey( props, bsStrDup(buffer), "video_size",    BSDICTFREE_VALUE );
                //    bsDictSetValueForKey( props, "mpeg2video",  "video_codec",      BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "2000000",     "video_bitrate",    BSDICTFREE_NONE );

				double fps = -1;
				if( pisr_NonRef!=NULL )
					fps = isrGetFPS( pisr_NonRef);
				else if( pisr_Ref!=NULL )
					fps = isrGetFPS( pisr_Ref );

				if(fps<=0)
				{
					fprintf(stderr, "Warning setting fps to 25 by default\n");
					fps = 25;
				}
				sprintf( buffer, "%f", fps );
                bsDictSetValueForKey( props, bsStrDup(buffer),          "video_fps",        BSDICTFREE_NONE );

                if( psb!= NULL )
                {
                    sprintf( buffer, "%d", psb->properties.sampleRate );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_rate", BSDICTFREE_VALUE );
                    
                    sprintf( buffer, "%d", psb->properties.numChannels );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_num_channels", BSDICTFREE_VALUE );
                    
                    bsDictSetValueForKey( props, "128000", "audio_bitrate", BSDICTFREE_NONE );
                    
                }
                
                sprintf( buffer, "%s/%s.mpg", argv[2], parts[0] );
                pisw = iswOpen( buffer, props );
                bsDictFree( props );
            }
            
            iswPutFrame( pisw, pimg, psb );
        }
        
        BSFREEFCT( pisw, iswClose );
        
        printf("\n");
        
        BSFREEFCT( pimg_NonRef, imgFree );
        BSFREEFCT( pimg_Ref,    imgFree );
        BSFREEFCT( pimg_Out,    imgFree );
        
        BSFREEFCT( psb_NonRef,  sndSBFree );
        BSFREEFCT( psb_Ref,     sndSBFree );
        
        BSFREEFCT( pisr_NonRef, isrClose );
        BSFREEFCT( pisr_Ref,    isrClose );

    }
    
    fclose( file );
    return 0;
}
