/*
 *  splitMovies.c
 *  dev
 *
 *  Created by laurent joyeux on 01/04/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */

#include <BaseLib/BaseLib.h>
#include <VideoLib/VideoLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

int main( int argc, char *argv[] )
{
    FILE    *file;
    
    if( argc<3 )
    {
        fprintf( stderr, "%s mergeMovie mergeTC [offset] [slope] [border]\n", argv[0] );
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
        
        return 1;
    }

    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");

    Image_t *pimg;
    ImageSequenceWriter_t *pisw = NULL;
    SoundBuffer_t *psb = NULL;
    int	offset = 0;
	float slope = 0;
	int	border = 0;
	
    if( argc>3 )
		offset = atoi(argv[3]);
	if( argc>4 )
		slope = atof(argv[4]);
	if( argc>5 )
		border = atoi(argv[4]);
		
    file = fopen(argv[2], "r");
    
    ImageSequenceReader_t *pisr = isrOpen(argv[1]);
    
    if( NULL==pisr )
    {
        fprintf(stderr, "Cannot open %s\n", argv[1]);
        abort();
    }
    
    int frame = 0;
    pimg = NULL;
    int movieCounter;
	for(movieCounter=0; ;movieCounter++)
    {
        char line[256];
        
        if( NULL==fgets(line, 256, file) )
            break;
        
        char *parts[3];
        char *src = line;
        int i;
        for( i=0; i<3; i++, src=NULL )
            parts[i] = strtok( src, " \t\n" );

        TimeCode_t tcIn, tcOut;
        tcStringToTC( &tcIn, parts[1] );
        tcStringToTC( &tcOut, parts[2] );

		int tcInFrame = tcIn.frameNumber-offset-slope*movieCounter;
		int tcOutFrame = tcOut.frameNumber-offset-slope*movieCounter;
		
        for(; frame<tcInFrame+border; frame++)
        {
            if( isrGetFrame( pisr, &pimg, NULL )!=0 )
                break;
        }
        
        if( frame<tcInFrame+border )
            break;
		
        
        for(; frame<tcOutFrame-border; frame++)
        {
            int result = isrGetFrame( pisr, &pimg, NULL );
            if( result!=0 )
                break;
            
            if( NULL==pisw )
            {
                char    buffer[256];
                BSDictionary_t *props;
                
                props = bsDictNew();

                sprintf( buffer, "%d,%d", imgGetWidth( pimg ), imgGetHeight( pimg ) );
                
                bsDictSetValueForKey( props, bsStrDup(buffer), "video_size",    BSDICTFREE_VALUE );
                //    bsDictSetValueForKey( props, "mpeg2video",  "video_codec",      BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "1000000",     "video_bitrate",    BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "25",          "video_fps",        BSDICTFREE_NONE );

                if( psb!= NULL )
                {
                    sprintf( buffer, "%d", psb->properties.sampleRate );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_rate", BSDICTFREE_VALUE );
                    
                    sprintf( buffer, "%d", psb->properties.numChannels );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_num_channels", BSDICTFREE_VALUE );
                    
                    bsDictSetValueForKey( props, "128000", "audio_bitrate", BSDICTFREE_NONE );
                    
                }
                
                pisw = iswOpen( parts[0], props );
                bsDictFree( props );
            }
            
            iswPutFrame( pisw, pimg, NULL );
        }
        
        BSFREEFCT( pisw, iswClose );
        BSFREEFCT( pimg, imgFree );
    }
    
    isrClose( pisr );
    
    
    return 0;
}
