/*
 *  splitMovies1.c
 *  dev
 *
 *  Created by laurent joyeux on 19/06/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */



#include <BaseLib/BaseLib.h>
#include <VideoLib/VideoLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <math.h>

double imgL2(const Image_t *pimg1, const Image_t *pimg2)
{
    uint8_t *pu8_img1, *pu8_img2;
    int     width, height;
    int     bytesPerRow;
    double  L2;
    int     i, j;
    
    width = imgGetWidth( pimg1 );
    height = imgGetHeight( pimg1 );
    bytesPerRow = imgGetBytesPerRow( pimg1 );
    
    pu8_img1 = imgGetTopLeftPixel( pimg1 );
    pu8_img2 = imgGetTopLeftPixel( pimg2 );
    
    L2 = 0;
    for(j=0; j<height; j++, pu8_img1+=bytesPerRow, pu8_img2+=bytesPerRow)
    {
        for(i=0; i<width; i++)
        {
            int diff;
            diff = pu8_img1[i*3+0]-pu8_img2[i*3+0];
            L2 += diff*diff;
            diff = pu8_img1[i*3+1]-pu8_img2[i*3+1];
            L2 += diff*diff;
            diff = pu8_img1[i*3+2]-pu8_img2[i*3+2];
            L2 += diff*diff;
        }
    }
    
    return sqrt(L2/(3*width*height));
}

int main( int argc, char *argv[] )
{
    FILE    *file;
    
    if( argc<3 )
    {
        fprintf( stderr, "%s mergeMovie mergeTC image_ref [offset] [slope] [border]\n", argv[0] );
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
        
        return 1;
    }

    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");

    Image_t *pimg;
    Image_t *pimg_ref;
    ImageSequenceWriter_t *pisw = NULL;
    SoundBuffer_t *psb = NULL;
    int	offset = 0;
	float slope = 0;
	int	border = 0;
	
    pimg_ref = imgRead( (int8_t*) argv[3] );
    if( NULL==pimg_ref )
        return -2;
        
    if( argc>4 )
		offset = atoi(argv[4]);
	if( argc>5 )
		slope = atof(argv[5]);
	if( argc>6 )
		border = atoi(argv[6]);
		
    file = fopen(argv[2], "r");
    
    ImageSequenceReader_t *pisr = isrOpen(argv[1]);
    
    if( NULL==pisr )
    {
        fprintf(stderr, "Cannot open %s\n", argv[1]);
        abort();
    }
    
    int frame = 0;
    pimg = NULL;
    int movieCounter;
	for(movieCounter=0; ;movieCounter++)
    {
        char line[256];
        
        if( NULL==fgets(line, 256, file) )
            break;
        
        char *parts[3];
        char *src = line;
        int i;
        for( i=0; i<3; i++, src=NULL )
            parts[i] = strtok( src, " \t\n" );

        TimeCode_t tcIn, tcOut;
        tcStringToTC( &tcIn, parts[1] );
        tcStringToTC( &tcOut, parts[2] );

		int tcInFrame = tcIn.frameNumber-offset-slope*movieCounter;
		int tcOutFrame = tcOut.frameNumber-offset-slope*movieCounter;
		
        for(; frame<tcInFrame+border; frame++)
        {
            if( isrGetFrame( pisr, &pimg, NULL )!=0 )
                break;
        }
        
        for(;;frame++)
        {
            if( isrGetFrame( pisr, &pimg, NULL )!=0 )
                break;
            
            double L2 = imgL2(pimg, pimg_ref);
            printf("frame %d L2 %f\n", frame, (float) L2);
            
            if( L2>5 )
                break;
        }
		
        
        for(; frame<tcOutFrame; frame++)
        {
            int result = isrGetFrame( pisr, &pimg, NULL );
            if( result!=0 )
                break;

            double L2 = imgL2(pimg, pimg_ref);
            printf("frame %d L2 %f\n", frame, (float) L2);
            
            if( L2<5 )
                break;
            
            if( NULL==pisw )
            {
                char    buffer[256];
                BSDictionary_t *props;
                
                props = bsDictNew();

                sprintf( buffer, "%d,%d", imgGetWidth( pimg ), imgGetHeight( pimg ) );
                
                bsDictSetValueForKey( props, bsStrDup(buffer), "video_size",    BSDICTFREE_VALUE );
                //    bsDictSetValueForKey( props, "mpeg2video",  "video_codec",      BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "1000000",     "video_bitrate",    BSDICTFREE_NONE );
                bsDictSetValueForKey( props, "25",          "video_fps",        BSDICTFREE_NONE );

                if( psb!= NULL )
                {
                    sprintf( buffer, "%d", psb->properties.sampleRate );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_rate", BSDICTFREE_VALUE );
                    
                    sprintf( buffer, "%d", psb->properties.numChannels );
                    bsDictSetValueForKey( props, bsStrDup(buffer), "audio_num_channels", BSDICTFREE_VALUE );
                    
                    bsDictSetValueForKey( props, "128000", "audio_bitrate", BSDICTFREE_NONE );
                    
                }
                
                pisw = iswOpen( parts[0], props );
                bsDictFree( props );
            }
            
            iswPutFrame( pisw, pimg, NULL );
        }
        
        BSFREEFCT( pisw, iswClose );
        BSFREEFCT( pimg, imgFree );
    }
    
    isrClose( pisr );
    
    
    return 0;
}
