/*
 *  transform.c
 *  dev
 *
 *  Created by laurent joyeux on 21/04/08.
 *  Copyright 2008 Laurent Joyeux. All rights reserved.
 *
 */

#include <ImageLib/ImageLib.h>
#include <VideoLib/VideoLib.h>
#include <SoundLib/SoundLib.h>
#include <math.h>

#include <ft2build.h>
#include FT_FREETYPE_H

#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/time.h>




/*
Traitement d'image en flottant compris entre 0 et 1, ARGB
*/

unsigned char clipUINT8(int a) 
{
	if(a<0)
		return 0;
	else if(a>255)
		return 255;
	else return a;
}

float clipFloat32(float a) 
{
	if(a<0)
		return 0;
	else if(a>1)
		return 1;
	else return a;
}

typedef struct
{
    float h, s, v;
} HSVC;

typedef struct
{
    float r, g, b;
} RGBC;


void RGBC2HSVC(const RGBC *in, HSVC *out)
{
    float max;
    float min, delta;
	
    out->h = out->s = out->v = 0;
    
    max = (in->r>in->g)? in->r:in->g;
    out->v = max = (max>in->b)? max : in->b;
	
    min = (in->r>in->g)? in->g:in->r;
    min = (min>in->b)? in->b : min;
	
    delta = max-min;
    if(max!=0)
        out->s = delta/max;
    else{
        out->s = 0;
        out->h = -1;
        return;
    }
	
    if(delta==0) return;
    
    if(in->r==max){
        out->h = (in->g - in->b)/delta;
    } else if (in->g==max){
        out->h = 2+(in->b - in->r)/delta;
    } else{
        out->h = 4+(in->r - in->g)/delta;
    }
    out->h *= 60;
    if(out->h<0) out->h += 360;
}

void HSVC2RGBC(const HSVC *in, RGBC *out)
{
	
    out->r = out->g = out->b = 0;
	
	if( 0==in->s )
	{
		out->r = out->g = out->b = in->v;
	}
	else
	{
		float	var_h, var_i, var_1, var_2, var_3;
		float	var_r, var_g, var_b;
		
		var_h = in->h * 6/360;
		if ( var_h == 6 ) 
			var_h = 0;      //H must be < 1
		var_i = floor( var_h );             //Or ... var_i = floor( var_h )

		var_1 = in->v * ( 1 - in->s );
		var_2 = in->v * ( 1 - in->s * ( var_h - var_i ) );
		var_3 = in->v * ( 1 - in->s * ( 1 - ( var_h - var_i ) ) );
		
		if ( var_i == 0 ) 
		{ 
			var_r = in->v; 
			var_g = var_3; 
			var_b = var_1; 
		}
		else if ( var_i == 1 ) 
		{ 
			var_r = var_2; 
			var_g = in->v; 
			var_b = var_1; 
		}
		else if ( var_i == 2 ) 
		{ 
			var_r = var_1; 
			var_g = in->v; 
			var_b = var_3; 
		}
		else if ( var_i == 3 ) 
		{ 
			var_r = var_1; 
			var_g = var_2; 
			var_b = in->v;
		}
		else if ( var_i == 4 ) 
		{ 
			var_r = var_3; 
			var_g = var_1; 
			var_b = in->v;
		}
		else
		{ 
			var_r = in->v; 
			var_g = var_1; 
			var_b = var_2; 
		}
		
		out->r = var_r;                 //RGB results = 0 ÷ 255
		out->g = var_g;
		out->b = var_b;
			
	}
}

typedef struct
{
	int				frameNumber;
    Image_t         *pimg;
    SoundBuffer_t   *psb;
} VideoFrame_t;

typedef struct  
{
    Image_t **ppimgs;
    int     length;
    
    int     offsetX, offsetY;
    float   speedX, speedY;
    
    int     pos;  // 0: left, 1: center, 2: right
    
} ImageWord_t;

typedef struct
{
    float   sigma;
} BlurEffect_t;

typedef struct
{
    float   gamma[3];
} GammaEffect_t;

typedef struct
{
    float   brightness;
    float   contrast;
} BrightnessAndContrastEffect_t;

typedef struct
{
    float hue, saturation, value;
} HSVEffect_t;

typedef struct
{
    int     kind[4];
    float   stddev[4];
} NoiseEffect_t;

typedef struct
{
    int offsetX, offsetY, width, height;
} CropEffect_t;

typedef struct
{
    int axis;
} FlipEffect_t;

typedef struct
{
    Image_t *pimg;
    ImageSequenceReader_t *pisr;
    int     frameNumber;
	int		startFrameNumber;
	int		endFrameNumber;
    
    int     offsetX, offsetY;
    int     width, height;
    
    float   scaleX, scaleY;
    float   alpha;
} PictureInPictureEffect_t;

typedef struct
{
    int offsetX, offsetY;
} ShiftEffect_t;

typedef struct
{
    ImageWord_t *piw;
    int         frameNumber;
} TextEffect_t;

typedef struct
{
    float scaleX, scaleY;
} ScaleEffect_t;

typedef struct
{
    float frequency;
} DropEffect_t;

typedef struct
{
    int factor;
    int width, height;
} ResampleEffect_t;

Image_t* imgRGBUint8ToARGBFloat32(const Image_t *pimg_In, Image_t **ppimg_Out);

void imgBlur(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, BlurEffect_t  *blur)
{    
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    float   gauss[256];
    int		size;
    float   *fSrc, *fDst, *src1;
    float	*fblurImage;
    int		doInv;
    float	sumGauss;
    int     width, height;
    int     numComp;
    int     i, j, k;
    float   sigma;
    
    width   = imgGetWidth( pimg_In );
    height  = imgGetHeight( pimg_In );
    numComp = imgFormatGetNumComp( imgGetFormat( pimg_In ) );
    
    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
    
    fblurImage = malloc(width*height*numComp*sizeof(float));
    
    sigma = blur->sigma;
    
    size = ceil(sigma)*3;
    if(size>127)
        size = 127;
    
    for(i=0; i<256; i++)
        gauss[i] = 0;
    
    sumGauss = 0;
    for(i=-size; i<=size; i++)
    {
        float d = i/sigma;
        sumGauss += gauss[127+i] = exp(-d*d/2)/(sqrt(2*M_PI)*sigma);
    }
    
    sumGauss = 1/sumGauss;
    for(i=-size; i<=size; i++)
    {
        gauss[127+i] *= sumGauss;
    }
    
    fDst = fblurImage;
    src1 = imgGetTopLeftPixel( pimg_In );
    int bytesPerRow = imgGetBytesPerRow( pimg_In );
    
    // X filter
    for(j=0; j<height; j++)
    {
        
        for(i=0; i<width; i++)
        {
            int	ib, ie, ii, iii;
            ib = i-size;
            ie = i+size;

            doInv = 0;
            iii = -size;
            if( ib<0 )
            {
                ib = 0;
                iii -= ib;
                doInv = 1;
            }

            if( ie>=width )
            {
                ie = width-1;
                doInv = 1;
            }
            float sum[5];
            sum[0] = sum[1] = sum[2] = sum[3] = sumGauss = 0;
            iii += 127;
            fSrc = (float*) ((char*) src1 +j*bytesPerRow) + ib*numComp;
            for(ii=ib; ii<=ie; ii++, iii++, fSrc+=numComp)
            {
                for(k=0; k<numComp; k++)
                    sum[k] += fSrc[k]*gauss[iii];

                sumGauss += gauss[iii];
            }
            
            float inv = doInv ? 1/sumGauss : 1;
            for(k=0; k<numComp; k++)
                *fDst++ = sum[k]*inv;
        }
    }

    // Y filter
    fDst = imgGetTopLeftPixel( *ppimg_Out );
    for(j=0; j<height; j++, fDst = (float*) ((char*)fDst + bytesPerRow) )
    {
        int	jb, je, jj, jjj;
        jb = j-size;
        je = j+size;
        
        doInv = 0;
        jjj = -size;
        if( jb<0 )
        {
            jb = 0;
            jjj -= jb;
            doInv = 1;
        }
        
        if( je>=height )
        {
            je = height-1;
            doInv = 1;
        }
        
        jjj += 127;
        for(i=0; i<width; i++)
        {
            float sum[5];
            int	jjjj;
            
            sum[0] = sum[1] = sum[2] = sum[3] = sumGauss = 0;
            float *fSrc = fblurImage + (i+jb*width)*numComp;
            
            jjjj = jjj;
            for(jj=jb; jj<=je; jj++, jjjj++, fSrc+=numComp*width)
            {
                for(k=0; k<numComp; k++)
                    sum[k] += fSrc[k]*gauss[jjjj];

                sumGauss += gauss[jjjj];
            }
            
            float inv = doInv ? 1/sumGauss : 1;
            for(k=0; k<numComp; k++)
                fDst[numComp*i+k] = sum[k]*inv;
        }
    }
    
    free( fblurImage );
}

void imgGamma(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, GammaEffect_t *gammaEffect)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    float *pf_Src, *pf_Dst;
    
    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
    
    pf_Dst = imgGetTopLeftPixel( *ppimg_Out );
    pf_Src = imgGetTopLeftPixel( pimg_In );
    
    int size = imgGetWidth( pimg_In ) * imgGetHeight( pimg_In ) ;
    int loop;
    
    float *gamma = gammaEffect->gamma;
    for(loop=0; loop<size; loop++)
    {
        *pf_Dst++ = *pf_Src++;
        *pf_Dst++ = powf(*pf_Src++, gamma[0]);
        *pf_Dst++ = powf(*pf_Src++, gamma[1]);
        *pf_Dst++ = powf(*pf_Src++, gamma[2]);
    }
}

void imgBrightnessAndContrast(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, BrightnessAndContrastEffect_t *bc)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    float *pf_Src, *pf_Dst;
    
    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
    
    pf_Dst = imgGetTopLeftPixel( *ppimg_Out );
    pf_Src = imgGetTopLeftPixel( pimg_In );
    
    float brightness = bc->brightness;
    float contrast = bc->contrast;
    
    int size = imgGetWidth( pimg_In ) * imgGetHeight( pimg_In ) * imgFormatGetNumComp( imgGetFormat( pimg_In ) );
    int loop;
    
    for(loop=0; loop<size; loop++)
        *pf_Dst++ = (*pf_Src++-0.5)*(contrast+1)+0.5 + brightness;
}

void imgHSV(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, HSVEffect_t *hsv)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    float *pf_Src, *pf_Dst;
    
    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
    
    float hue = hsv->hue;
    float saturation = hsv->saturation;
    float value = hsv->value;
    
    pf_Dst = imgGetTopLeftPixel( *ppimg_Out );
    pf_Src = imgGetTopLeftPixel( pimg_In );
    
    int size = imgGetWidth( pimg_In ) * imgGetHeight( pimg_In );
    int loop;
    
    for(loop=0; loop<size; loop++)
    {
        RGBC rgb;
        HSVC hsv;

        *pf_Dst++ = *pf_Src++;

        rgb.r = *pf_Src++; 
        rgb.g = *pf_Src++; 
        rgb.b = *pf_Src++;
        
        RGBC2HSVC(&rgb, &hsv);
        
        hsv.s *= saturation+1;
        hsv.s = clipFloat32( hsv.s );
        
        hsv.h += hue;
        if( hsv.h>=360 )
            hsv.h -= 360;
            
        hsv.v *= value+1;
        hsv.v = clipFloat32( hsv.v );
        
        
        HSVC2RGBC(&hsv, &rgb);
        
        *pf_Dst++ = rgb.r;
        *pf_Dst++ = rgb.g;
        *pf_Dst++ = rgb.b;
    }
}

/*
float ratio, 

    int incrustWidth  = imgGetWidth(pimg_Incrust)*ratio;
    int incrustHeight = imgGetHeight(pimg_Incrust)*ratio;
     
    Image_t *pimg_Tmp = imgInit( incrustWidth, incrustHeight, imgGetFormat(pimg_Incrust), imgGetType( pimg_Incrust) );
    
    imgResizeGeneric(pimg_Incrust, pimg_Tmp, IMGRESIZE_BILINEAR);

    BSFREEFCT( pimg_Tmp, imgFree );

*/

/*
kind:
    - 0 none
    - 1 white
    - 2 gaussian
*/

void imgNoise(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, NoiseEffect_t *noise)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    float *pf_Src, *pf_Dst;
    
    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
    
    pf_Dst = imgGetTopLeftPixel( *ppimg_Out );
    pf_Src = imgGetTopLeftPixel( pimg_In );
    
    int     *kind = noise->kind;
    float   *sigma = noise->stddev;
    
    int size = imgGetWidth( pimg_In ) * imgGetHeight( pimg_In );
    int loop, i, k;
    
    for(loop=0; loop<size; loop++)
    {
        for(i=0; i<4; i++)
        {
            float noise = 0;
            
            switch(kind[i])
            {
                case 0: noise = 0; break;
                case 1: noise = (drand48()-0.5)*sigma[i]*2; break;
                case 2:
                    for(noise=0, k=0; k<6; k++)
                        noise += drand48()-0.5;
                    noise *= sigma[i]/0.7069;
                    
                    break;
            }
            
            *pf_Dst++ = *pf_Src++ + noise;
        }
    }
}

void imgCrop(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, CropEffect_t *crop)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    imgCopy( pimg_In, ppimg_Out );
    
    int offsetX = crop->offsetX;
    int offsetY = crop->offsetY;
    int width   = crop->width;
    int height  = crop->height;
    
    int inWidth  = imgGetWidth( pimg_In );
    int inHeight = imgGetHeight( pimg_In );
    
    if( offsetX>inWidth || offsetX+width<0 )
        return;

    if( offsetY>inHeight || offsetY+height<0 )
        return;
    
    float *pf_Dst;

    pf_Dst = imgGetTopLeftPixel( *ppimg_Out );
    
    int bytesPerRow = imgGetBytesPerRow( pimg_In );
    
    int i, j;
    for(j=0; j<inHeight; j++, pf_Dst = (float*) ((char*)pf_Dst + bytesPerRow) )
    {
        int clearY = j<offsetY || j>=offsetY+height ;

        for(i=0; i<inWidth; i++)
        {
            int clearX = i<offsetX || i>=offsetX+width;
            
            if(clearX || clearY)
            {
                pf_Dst[4*i+0] = 0;
                pf_Dst[4*i+1] = 0;
                pf_Dst[4*i+2] = 0;
                pf_Dst[4*i+3] = 0;
            }
        }
    }
}

void imgFlip(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, FlipEffect_t *flip)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    if( NULL == *ppimg_Out )
        *ppimg_Out = imgInitWithImage( pimg_In );
        
    float *pf_In, *pf_Out;
    
    pf_In  = (float*) imgGetTopLeftPixel( pimg_In );
    pf_Out = (float*) imgGetTopLeftPixel( *ppimg_Out );

    int axis = flip->axis;

    int bytesPerRow = imgGetBytesPerRow( pimg_In );
    int width       = imgGetWidth( pimg_In );
    int height      = imgGetHeight( pimg_In );
    
    int i, j;
    
    if( 0==axis )
    {
        pf_Out = (float*) ((height-1)*bytesPerRow + (char *) pf_Out);
        
        for(j=0; j<height; j++, pf_In=(float*) ((char*)pf_In+bytesPerRow), pf_Out=(float*) ((char*)pf_Out-bytesPerRow))
        {
            for(i=0; i<width*4; i++)
            {
                pf_Out[i] = pf_In[i];
            }
        }
    }
    else
    {
        for(j=0; j<height; j++, pf_In=(float*) ((char*)pf_In+bytesPerRow), pf_Out=(float*) ((char*)pf_Out+bytesPerRow))
        {
            for(i=0; i<width; i++)
            {
                pf_Out[(width-1-i)*4+0] = pf_In[4*i+0];
                pf_Out[(width-1-i)*4+1] = pf_In[4*i+1];
                pf_Out[(width-1-i)*4+2] = pf_In[4*i+2];
                pf_Out[(width-1-i)*4+3] = pf_In[4*i+3];
            }
        }
    }
}

void imgPictureInPictureSub(const Image_t *pimg_In, const Image_t *pimg_Incrust, Image_t **ppimg_Out, int offsetX, int offsetY )
{
    imgCopy( pimg_In, ppimg_Out );
    
    int incrustWidth  = imgGetWidth( pimg_Incrust );
    int incrustHeight = imgGetHeight( pimg_Incrust );

    int inWidth  = imgGetWidth( pimg_In );
    int inHeight = imgGetHeight( pimg_In );
    
    int incrustOffsetX, incrustOffsetY;
    
    
    if( offsetX>inWidth )
        return;
    
    int width;
    
    if( offsetX<0 )
    {
        incrustOffsetX = -offsetX;
        if( incrustOffsetX > incrustWidth )
            return;
        offsetX = 0;
        
        width = incrustWidth - incrustOffsetX;
        if( width > inWidth )
             width = inWidth;
    }
    else
    {
        incrustOffsetX = 0;

        width = inWidth - offsetX;
        if( width > incrustWidth )
            width = incrustWidth;
    }
    
    if( width < 1 )
        return ;


    if( offsetY>inHeight )
        return;
    
    int height;
    
    if( offsetY<0 )
    {
        incrustOffsetY = -offsetY;
        if( incrustOffsetY > incrustHeight )
            return;
        offsetY = 0;
        
        height = incrustHeight - incrustOffsetY;
        if( height > inHeight )
             height = inHeight;
    }
    else
    {
        incrustOffsetY = 0;

        height = inHeight - offsetY;
        if( height > incrustHeight )
            height = incrustHeight;
    }
    
    if( height < 1 )
        return ;
    
    int incrustBytesPerRow = imgGetBytesPerRow( pimg_Incrust );
    int inBytesPerRow = imgGetBytesPerRow( pimg_In );

    float *pf_Out, *pf_Incrust;

    pf_Out = (float*) imgGetTopLeftPixel( *ppimg_Out );
    pf_Incrust = (float*) imgGetTopLeftPixel( pimg_Incrust );
    
    pf_Out = (float*) ( (char*) pf_Out + offsetY*inBytesPerRow) + offsetX*4;

    pf_Incrust = (float*) ( (char*) pf_Incrust + incrustOffsetY*incrustBytesPerRow) + incrustOffsetX*4;
    
    int i, j;
    for(j=0; j<height; j++, pf_Out=(float *)((char *)pf_Out + inBytesPerRow), pf_Incrust=(float *)((char *)pf_Incrust+incrustBytesPerRow) )
    {
        for(i=0; i<width; i++)
        {
            float alpha = pf_Incrust[4*i+0];
            
            pf_Out[4*i+0] = 1; //alpha;
            pf_Out[4*i+1] = pf_Out[4*i+1]*(1-alpha) + pf_Incrust[4*i+1]*alpha;
            pf_Out[4*i+2] = pf_Out[4*i+2]*(1-alpha) + pf_Incrust[4*i+2]*alpha;
            pf_Out[4*i+3] = pf_Out[4*i+3]*(1-alpha) + pf_Incrust[4*i+3]*alpha;
        }
    }    
}

void imgPictureInPicture(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, PictureInPictureEffect_t *picture)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    Image_t     *pimg = picture->pimg;
    ImageSequenceReader_t *pisr = picture->pisr;
    Image_t     *pimg_isr = NULL;
    Image_t     *pimg_Tmp = NULL;
    Image_t     *pimg_Src;
    int			inInterval = (pvf_In->frameNumber>=picture->startFrameNumber) && (picture->endFrameNumber==-1 || pvf_In->frameNumber<picture->endFrameNumber);
	
    if( pimg!=NULL  && inInterval)
    {
        pimg_Src = pimg;
    }
    else if( pisr!=NULL && inInterval)
    {
		SoundBuffer_t *psb = NULL;
        if(isrGetFrame( pisr, & pimg_isr, &psb )<0)
        {
            isrClose( pisr );
            picture->pisr = NULL;

            imgCopy(pimg_In, ppimg_Out);
            
            return;
        }
        
        imgRGBUint8ToARGBFloat32( pimg_isr, &pimg_Tmp );
        if(psb!=NULL)
		{
			if(   pvf_In->psb->properties.sampleRate!=psb->properties.sampleRate 
			   || pvf_In->psb->properties.numChannels!=psb->properties.numChannels
			   || pvf_In->psb->properties.format!=psb->properties.format)
			{
				SoundBuffer_t *psb_converted = sndSBInit( & pvf_In->psb->properties, 0);
				if(sndSBConvert(psb, psb_converted)!=0)
				{
					fprintf(stderr, "%s:%d sound convertion failed %d %d, %d %d, %d %d\n", 
							__FILE__, __LINE__,
							pvf_In->psb->properties.sampleRate, psb->properties.sampleRate,
							pvf_In->psb->properties.numChannels, psb->properties.numChannels,
							pvf_In->psb->properties.format, psb->properties.format);
					abort();
				}
				sndSBFree(psb);
				psb = psb_converted;
			}
			BSFREEFCT(pvf_Out->psb, sndSBFree);
			pvf_Out->psb = psb;
		}
        imgFree( pimg_isr );
        pimg_isr = pimg_Tmp;
        pimg_Tmp = NULL;
        
        pimg_Src = pimg_isr;
        picture->frameNumber++;
    }
    else
    {
        imgCopy(pimg_In, ppimg_Out);
        return;
    }
    
    
    if( picture->scaleX!=1 || picture->scaleY!=1 )
    {
        int srcWidth  = imgGetWidth( pimg_Src ) * picture->scaleX;
        int srcHeight = imgGetHeight( pimg_Src ) * picture->scaleY;
        
        pimg_Tmp = imgInit( srcWidth, srcHeight, IMGFORMAT_ARGB, FLOAT32_TYPE );
        imgResize( pimg_Src, pimg_Tmp, IMGRESIZE_BILINEAR );
    }
    
    imgPictureInPictureSub( pimg_In, pimg_Tmp ? pimg_Tmp : pimg_Src, ppimg_Out, picture->offsetX, picture->offsetY );
    
    BSFREEFCT( pimg_isr, imgFree );
    BSFREEFCT( pimg_Tmp, imgFree );
}

void imgShift(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, ShiftEffect_t *shift)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    Image_t *pimg_Zero = imgInitWithImage( pimg_In );
    
    int offsetX = shift->offsetX;
    int offsetY = shift->offsetY;
     
    float *pf = imgGetBuffer( pimg_Zero );
    int numPixels = imgGetWidth( pimg_In )*imgGetHeight( pimg_In );
    
    int loop;
    for(loop=0; loop<numPixels; loop++)
    {
        *pf++ = 0;
        *pf++ = 0;
        *pf++ = 0;
        *pf++ = 0;
    }
     
    imgPictureInPictureSub( pimg_Zero, pimg_In, ppimg_Out, offsetX, offsetY );
    
    imgFree( pimg_Zero );
}

void imgScale(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, ScaleEffect_t *scale)
{    
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    int scaledWidth = imgGetWidth( pimg_In )*scale->scaleX;
    int scaledHeight = imgGetHeight( pimg_In )*scale->scaleY;
    Image_t *pimg = imgInit( scaledWidth,  scaledHeight, IMGFORMAT_ARGB, FLOAT32_TYPE );
    
    imgResize( pimg_In, pimg, IMGRESIZE_BILINEAR );
    
    int width = imgGetWidth(pimg_In);
    int height = imgGetHeight(pimg_In);
    
    int offsetX = (width - scaledWidth)/2;
    int offsetY = (height - scaledHeight)/2;
    
    Image_t *pimg_Zero = imgInitWithImage( pimg_In );
    
    float *pf = imgGetBuffer( pimg_Zero );
    int numPixels = imgGetWidth( pimg_In )*imgGetHeight( pimg_In );
    
    int loop;
    for(loop=0; loop<numPixels; loop++)
    {
        *pf++ = 0;
        *pf++ = 0;
        *pf++ = 0;
        *pf++ = 0;
    }

    imgPictureInPictureSub( pimg_Zero, pimg, ppimg_Out, offsetX, offsetY );

    imgFree( pimg_Zero );    
    imgFree( pimg );
}

void imgDrop(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, DropEffect_t *drop)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    int dropFlag = drand48()<drop->frequency;
    
    if( dropFlag )
    {
        Image_t *pimg_Zero = imgInitWithImage( pimg_In );
        imgCopy(pimg_Zero, ppimg_Out);
        imgFree( pimg_Zero );
        if( pvf_In->psb!=NULL )
        {
            if( NULL == pvf_Out->psb )
                pvf_Out->psb = sndSBCopy( pvf_In->psb );
                
            memset(pvf_Out->psb->data, 0, pvf_Out->psb->allocatedBufferSize);
        }
    }
    else
        imgCopy(pimg_In, ppimg_Out);
    
}

void imgResample(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, ResampleEffect_t *resample)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    int sampleWidth;  
    int sampleHeight; 
    
    if( resample->factor!=0 )
    {
        sampleWidth  = imgGetWidth( pimg_In )/resample->factor;
        sampleHeight = imgGetHeight( pimg_In )/resample->factor;
    }
    else
    {
        sampleWidth  = resample->width;
        sampleHeight = resample->height;
    }
            
    *ppimg_Out = imgInit(sampleWidth, sampleHeight, IMGFORMAT_ARGB, FLOAT32_TYPE);
    if(resample->factor>1)
        imgResize(pimg_In, *ppimg_Out, IMGRESIZE_NN);
    else
        imgResize(pimg_In, *ppimg_Out, IMGRESIZE_BILINEAR);
}

void imgDrawTrueType( const FT_Face face, Image_t **pimg, float fg[4], float bg[4] )
{
    FT_Int          i, j;
    float           *pf;
    FT_Bitmap       *bitmap;
    int             height;
    FT_GlyphSlot    slot;
    
    slot = face->glyph;
    
    int width = slot->advance.x/64;
    bitmap = & slot->bitmap;

    FT_Int  x_max = bitmap->width;
    FT_Int  y_max = bitmap->rows;
    
    if( NULL == *pimg )
        *pimg = imgInit( width, face->size->metrics.height/64, IMGFORMAT_ARGB, FLOAT32_TYPE);
    
    pf = imgGetTopLeftPixel( *pimg );
    height = imgGetHeight( *pimg );
         
    for ( j = 0; j < height; j++ )
    {
        int offsetJ = (face->size->metrics.ascender-slot->metrics.horiBearingY)/64; //height-slot->bitmap_top;
        int jj = j-offsetJ;
        
        for ( i = 0; i < width; i++ )
        {
            int value = 0;
            int ii = i-slot->metrics.horiBearingX/64;
            
            if( (ii>=0 && ii<x_max) && (jj>=0 && jj<y_max) )
                value = bitmap->buffer[jj * bitmap->width + ii];
                
            float *color = (value>0) ? fg : bg;

            *pf++ = color[0];
            *pf++ = color[1];
            *pf++ = color[2];
            *pf++ = color[3];
        }
    }
}

#pragma mark ---

Image_t* imgBlank(int width, int height, float bg[4])
{
    Image_t *pimg = imgInit(width, height, IMGFORMAT_ARGB, FLOAT32_TYPE);
    
    float   *pf = imgGetBuffer( pimg );
    int     numPixels = width*height;
    
    int     loop;
    for(loop=0; loop<numPixels; loop++)
    {
        *pf++ = bg[0];
        *pf++ = bg[1];
        *pf++ = bg[2];
        *pf++ = bg[3];
    }
    
    return pimg;
}

ImageWord_t* iwInit(char *word, int size, FT_Face face, float fg[4], float bg[4])
{
    int length = strlen(word);
    
    ImageWord_t *iw = bsCalloc( sizeof(ImageWord_t) );
    iw->ppimgs = bsCalloc( length*sizeof(Image_t*) );
    iw->length = length;
    
    int error = FT_Set_Pixel_Sizes(
            face,   /* handle to face object */
            0,      /* pixel_width           */
            size );
            
    
    int n;
    for(n=0; word[n]!='\0'; n++)
    {
        if( ' '==word[n] )
        {
            iw->ppimgs[n] = imgBlank(face->size->metrics.max_advance/64/4, face->size->metrics.height/64, bg);
            continue;
        }
            
        error = FT_Load_Char( face, word[n], FT_LOAD_RENDER );
        if ( error )
          continue; 

        imgDrawTrueType( face, iw->ppimgs+n, fg, bg );
    }
   
    return iw;
}

void iwFree( ImageWord_t *piw )
{
    int loop;
    
    for(loop=0; loop<piw->length; loop++)
        BSFREEFCT( piw->ppimgs[loop], imgFree );
    
	BSFREE( piw->ppimgs );
	
    bsFree( piw );
}

#pragma mark ---

void imgText(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, TextEffect_t *text)
{
    Image_t *pimg_In    = pvf_In->pimg;
    Image_t **ppimg_Out  = & pvf_Out->pimg;

    if( NULL == *ppimg_Out )
        imgCopy( pimg_In, ppimg_Out );
    
    ImageWord_t *iw = text->piw;
    int frameNumber = text->frameNumber;
    
    int posX, posY;
    
    posX = (int) (iw->offsetX+iw->speedX*frameNumber);
    posY = (int) (iw->offsetY+iw->speedY*frameNumber);
    
    int loop;

    switch (iw->pos) 
    {
        case 0:
            break;
        case 1:
        {
            int sumX;
            
            for(loop=0, sumX=0; loop<iw->length; loop++)
                sumX += imgGetWidth( iw->ppimgs[loop] );
            
            posX -= sumX/2;
            
            break;
        }
        case 2:
        {
            int sumX;
            
            for(loop=0, sumX=0; loop<iw->length; loop++)
                sumX += imgGetWidth( iw->ppimgs[loop] );
            
            posX -= sumX;

            break;
        }
        default:
            break;
    }

    for(loop=0; loop<iw->length; loop++)
    {
        imgPictureInPictureSub( *ppimg_Out, iw->ppimgs[loop], ppimg_Out, posX, posY );
        posX += imgGetWidth( iw->ppimgs[loop] );
    }
    
    text->frameNumber++;
}

Image_t* imgRGBUint8ToARGBFloat32(const Image_t *pimg_In, Image_t **ppimg_Out)
{
    if( NULL==*ppimg_Out )
        *ppimg_Out = imgInit(imgGetWidth(pimg_In), imgGetHeight(pimg_In), IMGFORMAT_ARGB, FLOAT32_TYPE);

    imgConvert( pimg_In, *ppimg_Out, NULL );
    
    float *pf = imgGetTopLeftPixel( *ppimg_Out );
    int numPixels = imgGetWidth( *ppimg_Out )*imgGetHeight( *ppimg_Out );
    int loop;
    int numComp = imgFormatGetNumComp( imgGetFormat( pimg_In ) );
    
    if( 3==numComp )
    {
        for(loop=0; loop<numPixels; loop++)
        {
            *pf++  = 1;
            *pf++ *= 1/255.0f;
            *pf++ *= 1/255.0f;
            *pf++ *= 1/255.0f;
        }
    }
    else if( 4==numComp )
    {
        for(loop=0; loop<numPixels*numComp; loop++)
            *pf++ *= 1/255.0f;
    }
    
    return *ppimg_Out;
}

Image_t* imgARGBFloat32ToRGBUint8(const Image_t *pimg_In, Image_t **ppimg_Out)
{
    if( NULL==*ppimg_Out )
        *ppimg_Out = imgInit(imgGetWidth(pimg_In), imgGetHeight(pimg_In), IMGFORMAT_RGB, UINT8_TYPE);
    
    float *pf = imgGetTopLeftPixel( pimg_In );
    int size = imgGetWidth( pimg_In )*imgGetHeight( pimg_In )*4;
    int loop;
    
    for(loop=0; loop<size; loop++)
    {
        float value = (*pf)*255.0;

        if( value<0 )
            value = 0;
        if( value>255 )
            value = 255;

        *pf++ = value;
    }
    
    imgConvert( pimg_In, *ppimg_Out, NULL );
    
    return *ppimg_Out;
}

typedef void (*EffectFct_t)(const VideoFrame_t *pvf_In, VideoFrame_t *pvf_Out, void *effect);
typedef void (*EffectFree)(void *effect);

void freePictureInPicture(PictureInPictureEffect_t *picture)
{
    BSFREEFCT( picture->pisr, isrClose );
    BSFREEFCT( picture->pimg, imgFree );
}

void freeText(TextEffect_t *text)
{
    BSFREEFCT( text->piw, iwFree );
}

typedef struct
{
    void        *effect;
    EffectFct_t fct;
    EffectFree  free;
} Effect_t;

const char * findKey(const char **keys, const char **values, const char *key)
{
    while( *keys != NULL && strcmp(*keys, key) )
    {  
        keys++;
        values++;
    }
    
    return *values;
}

int main(int argc, char *argv[])
{
    FILE            *file;
    struct timeval  current_time;
    
    if( 1==argc )
    {
        fprintf(stderr, "%s file_effects \n", argv[0]);
        fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
        
        return -1;
    }
    
    fprintf(stderr, "(c) INRIA-IMEDIA 2008. This program was created for TrecVid2008\n");
    
    file = fopen(argv[1], "r");
    if( NULL==file )
        return -1;
    
    gettimeofday(&current_time, NULL);
    
    srand48(current_time.tv_sec);
    
    int eofFlag = 0;
    for(;!eofFlag;)
    {
        int         numEffects;
        Effect_t    *effects[100];
        int         videoRate;
        int         audioRate;
        TimeCode_t  tcIn, tcOut;
        int         tcInFlag = 0;
        int         tcOutFlag = 0;
        int         noSound = 0;
        
        int loop;
        for(loop=0; loop<100; loop++)
            effects[loop] = NULL;
        
        videoRate = audioRate = -1;
        
        numEffects = 0;
        char    *inputFile  = NULL;
        char    *outputFile = NULL;
        for(;;)
        {
            char  line[512];
            
            if( NULL==fgets(line, 512, file) )
            {
                eofFlag = 1;
                break;
            }
            printf("%s", line);
            
            if( 0==strlen( line ) )
                break;

            char    *src = line;
            
            while( *src!='\0' && isspace(*src) )
                src++;
            
            if( '#'==*src )
                continue;

            if( '\n'==*src || '\0'==*src )
                break;

            char    *keys[100];
            char    *values[100];
            int     nbParts;
            int     inQuote = 0;

            for(loop=0; loop<100; loop++)
                keys[loop] = values[loop] = NULL;

            nbParts = 0;
            keys[0] = src;
            for(; *src; src++)
            {
                if( inQuote && *src!='"' )
                    continue;
                
                if( '"'==*src )
                {
                    inQuote = !inQuote;
                }
                else if( isspace(*src) && !inQuote )
                {
                    *src = '\0';
                    src++;
                    while( *src && isspace( *src ) )
                        src++;
                    if( *src )
                    {
                        nbParts++;
                        keys[nbParts] = src;
                    }
					else
						break;
                }
                else if( '='==*src )
                {
                    *src = '\0';
                    values[nbParts] = src+1;
                }
            }
            
            char *cmd;
            
            cmd = "blur";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                // blur=amount
                BlurEffect_t *blur = bsCalloc( sizeof(BlurEffect_t ) );
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                
                const char *value = findKey((const char**)keys, (const char**)values, "blur");

                blur->sigma = atof(value);
                effect->fct = (void*) imgBlur;
                effect->effect = blur;
                effects[numEffects++] = effect;                

            }
            
            cmd = "gamma";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                // gamma red=x green=y blue=z
                GammaEffect_t *gamma = bsCalloc( sizeof(GammaEffect_t) );
                
                gamma->gamma[0] = 1;
                gamma->gamma[1] = 1;
                gamma->gamma[2] = 1;

                const char *value = findKey((const char**)keys, (const char**)values, "gamma");
                if( sscanf(value, "%f,%f,%f", &gamma->gamma[0], &gamma->gamma[1], &gamma->gamma[2])!=3 )
                {
                    fprintf(stderr, "unrecognized gamma %s\n", value);
                    abort();
                }
                
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) gamma;
                effect->fct = (void*) imgGamma;
                effects[numEffects++] = effect;
                
            }
            
            cmd = "level";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                BrightnessAndContrastEffect_t *bc = bsCalloc( sizeof(BrightnessAndContrastEffect_t) );
                
                const char *brightness = findKey((const char**)keys, (const char**)values, "brightness");
                if( brightness!=NULL )
                    bc->brightness = atof(brightness);

                const char *contrast = findKey((const char**)keys, (const char**)values, "contrast");
                if( contrast!=NULL )
                    bc->contrast = atof(contrast);

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) bc;
                effect->fct = (void*) imgBrightnessAndContrast;
                effects[numEffects++] = effect;

            }
            
            cmd = "color";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                HSVEffect_t *hsv = bsCalloc( sizeof(HSVEffect_t) );
                
                const char *hue = findKey((const char**)keys, (const char**)values, "hue");
                if( hue!=NULL )
                    hsv->hue = atof(hue);

                const char *value = findKey((const char**)keys, (const char**)values, "value");
                if( value!=NULL )
                    hsv->value = atof(value);

                const char *saturation = findKey((const char**)keys, (const char**)values, "saturation");
                if( saturation!=NULL )
                    hsv->saturation = atof(saturation);

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) hsv;
                effect->fct = (void*) imgHSV;
                effects[numEffects++] = effect;

            }
            
            cmd = "noise";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                NoiseEffect_t *noise =  bsCalloc( sizeof(NoiseEffect_t) );            

                const char *kind = findKey((const char**)keys, (const char**)values, "kind");
                if( kind!=NULL )
                {
                    // none,white,gaussian
                    int     kinds[4];
                    
                    int     affected;
                    int     i;
                    char    *src = (char *) kind;
                    for(affected=0; affected<4; src=NULL, affected++)
                    {
                        char *k = strtok( src, ",");
                        if( NULL==k )
                            break;
                            
                        if( !strcmp(k, "none" ) )
                            kinds[affected] = 0;
                        else if( !strcmp( k, "white") )
                            kinds[affected] = 1;
                        else if( !strcmp( k, "gaussian") )
                            kinds[affected] = 2;
                        else
                        {
                            fprintf(stderr, "unrecognized noise kind %s\n", k);
                            abort();
                        }
                    }
                                    
                    if( 3 == affected )
                    {
                        for(i=0; i<3; i++)
                            noise->kind[i+1] = kinds[i];
                    }
                    else if( 4 == affected )
                    {
                        for(i=0; i<4; i++)
                            noise->kind[i] = kinds[i];
                    }
                    else
                    {
                            fprintf(stderr, "need R G B kind of noise\n");
                            abort();
                    }
                    
                }

                const char *stddev = findKey((const char**)keys, (const char**)values, "stddev");
                if( stddev!=NULL )
                {
                    float dev[4];
                    int     i;
                    char    *src = (char *) stddev;
                    int     affected;

                    for(affected=0; affected<4; src=NULL, affected++)
                    {
                        char *k = strtok( src, ",");
                        if( NULL==k )
                            break;
                        if( sscanf(k, "%f", dev+affected)!=1 )
                        {
                            fprintf(stderr, "unrecognized noise value %s\n", k);
                            abort();
                        }
                    }
                    
                    if( 3 == affected )
                    {
                        for(i=0; i<3; i++)
                            noise->stddev[i+1] = dev[i];
                    }
                    else if( 4 == affected )
                    {
                        for(i=0; i<4; i++)
                            noise->stddev[i] = dev[i];
                    }
                    else
                    {
                            fprintf(stderr, "need R G B stddev of noise\n");
                            abort();
                    }
                    
               
                }

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) noise;
                effect->fct = (void*) imgNoise;
                effects[numEffects++] = effect;

            }

            cmd = "crop";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                CropEffect_t *crop = bsCalloc( sizeof(CropEffect_t) );
                       
                const char *offset = findKey((const char**)keys, (const char**)values, "offset");
                if( offset!=NULL )
                {
                    sscanf(offset, "%d,%d", & crop->offsetX, & crop->offsetY);
                }

                const char *size = findKey((const char**)keys, (const char**)values, "size");
                if( size!=NULL )
                {
                    sscanf(size, "%d,%d", & crop->width, & crop->height);
                }

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) crop;
                effect->fct = (void*) imgCrop;
                effects[numEffects++] = effect;
            }

            cmd = "flip";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                FlipEffect_t *flip = bsCalloc( sizeof(FlipEffect_t) );
                
                const char *value = findKey((const char**)keys, (const char**)values, "flip");
                if( value!=NULL)
                {
                    sscanf( value, "%d", & flip->axis );
                }

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) flip;
                effect->fct = (void*) imgFlip;
                effects[numEffects++] = effect;
            }
            
            cmd = "pictInPict";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                PictureInPictureEffect_t *pict = bsCalloc( sizeof(PictureInPictureEffect_t) );
                
                pict->alpha = 1;
                pict->scaleX = 1;
                pict->scaleY = 1;
				pict->startFrameNumber = 0;
				pict->endFrameNumber = -1;
                
                const char *image = findKey((const char**)keys, (const char**)values, "image");
                const char *movie = findKey((const char**)keys, (const char**)values, "movie");
                const char *scale = findKey((const char**)keys, (const char**)values, "scale");
                const char *tcStart = findKey((const char**)keys, (const char**)values, "tcStart");
                const char *tcEnd = findKey((const char**)keys, (const char**)values, "tcEnd");
                
                if( scale!=NULL )
                {
                    if( sscanf(scale, "%f,%f", &pict->scaleX, &pict->scaleY)!=2 )
                    {
                        fprintf(stderr, "unrecognized scale %s\n", scale);
                        abort();
                    }
                }
                
				if( tcStart!=NULL )
				{
					TimeCode_t tc;
					
					tcStringToTC(&tc, tcStart);
					pict->startFrameNumber = tc.frameNumber;
				}
				
				if( tcEnd!=NULL )
				{
					TimeCode_t tc;
					
					tcStringToTC(&tc, tcStart);
					pict->endFrameNumber = tc.frameNumber;
				}
				
                if( image!=NULL)
                {
                    
                    Image_t *pimg = imgRead((int8_t*) image);
                    if( NULL==pimg )
                    {
                        fprintf( stderr, "cannot read %s\n", image );
                        abort();
                    }
                    
                    if( pict->scaleX!=1 || pict->scaleY!=1 )
                    {
                        Image_t *pimg_Resized = imgInit(imgGetWidth(pimg)*pict->scaleX, imgGetHeight(pimg)*pict->scaleY, imgGetFormat(pimg), imgGetType(pimg));
                        imgResize(pimg, pimg_Resized, IMGRESIZE_BILINEAR);
                        
                        imgFree(pimg);
                        pimg = pimg_Resized;
                        
                        pict->scaleX = pict->scaleY = 1;
                    }
                    
                    imgRGBUint8ToARGBFloat32( pimg, & pict->pimg );                
                    imgFree( pimg );
                }
                else if( movie!=NULL )
                {
                    const char *alpha = findKey((const char**)keys, (const char**)values, "alpha");
                    //const char *alpha_audio = findKey((const char**)keys, (const char**)values, "alpha_audio");
                    const char *tcOffset  = findKey((const char**)keys, (const char**)values, "tcOffset");
                    //const char *video_bitrate  = findKey((const char**)keys, (const char**)values, "video_bitrate");
                    //const char *audio_bitrate  = findKey((const char**)keys, (const char**)values, "audio_bitrate");
                    
                    pict->pisr = isrOpen( movie );
                    if( NULL==pict->pisr )
                    {
                        fprintf( stderr, "cannot open %s\n", movie );
                        abort();
                    }
                    
                    if( alpha )
                        sscanf(alpha, "%f", & pict->alpha );
                    
                    if( tcOffset!=NULL )
                    {
                        TimeCode_t tc;
                        
                        tcStringToTC(&tc, tcOffset);
                        isrSeekToFrame( pict->pisr, (double)tc.frameNumber / tc.fps );
                    }
                }

                const char *offset = findKey((const char**)keys, (const char**)values, "offset");
                if( offset!=NULL )
                {
                    if( sscanf(offset, "%d,%d", &pict->offsetX, &pict->offsetY)!=2 )
                    {
                        fprintf(stderr, "unrecognized offset %s\n", offset);
                        abort();
                    }
                }
                
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect  = (void*) pict;
                effect->fct     = (void*) imgPictureInPicture;
                effect->free    = (void*) freePictureInPicture;
                effects[numEffects++] = effect;
            }
            
            cmd = "shift";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                ShiftEffect_t *shift = bsCalloc( sizeof(ShiftEffect_t) );
                
                const char *offset = findKey((const char**)keys, (const char**)values, "shift");
                if( offset!=NULL )
                {
                    sscanf(offset, "%d,%d", & shift->offsetX, & shift->offsetY);
                }
               
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect  = (void*) shift;
                effect->fct     = (void*) imgShift;
                effects[numEffects++] = effect;
            }

            cmd = "text";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                TextEffect_t *textEffect = bsCalloc( sizeof(TextEffect_t) );
                
                char *text = (char*) findKey((const char**)keys, (const char**)values, "text");
                if( text!=NULL )
                {
                    if( '"' == text[0] )
                    {
                        int loop;
                        for(loop=1; text[loop]!='\0'; loop++)
                            if( '"'==text[loop] )
                            {
                                text[loop] = '\0';
                                break;
                            }
                        text++;
                    }
                }
                
                float bg[4];
                bg[0] = 1; bg[1] = 0; bg[2] = 0; bg[3] = 0;
                const char *bgColor = findKey((const char**)keys, (const char**)values, "bg");
                if( bgColor!=NULL )
                {
                    if( sscanf( bgColor, "%f,%f,%f,%f", bg+0, bg+1, bg+2, bg+3 )!=4 )
                    {
                        sscanf( bgColor, "%f,%f,%f", bg+1, bg+2, bg+3 );
                        bg[0] = 1;
                    }
                }

                float fg[4];
                fg[0] = 1; fg[1] = 1; fg[2] = 1; fg[3] = 1;
                const char *fgColor = findKey((const char**)keys, (const char**)values, "fg");
                if( fgColor!=NULL )
                {
                    if( sscanf( fgColor, "%f,%f,%f,%f", fg+0, fg+1, fg+2, fg+3 )!=4 )
                    {
                        sscanf( fgColor, "%f,%f,%f", fg+1, fg+2, fg+3 );
                        fg[0] = 1;
                    }
                }
                
                int offset[2];
                offset[0] = 0; offset[1] = 0;
                const char *offsetStr = findKey((const char**)keys, (const char**)values, "offset");
                if( offsetStr!=NULL )
                {
                    sscanf( offsetStr, "%d,%d", offset+0, offset+1 );
                }

                float speed[2];
                speed[0] = 0; speed[1] = 0;
                const char *speedStr = findKey((const char**)keys, (const char**)values, "speed");
                if( speedStr!=NULL )
                {
                    sscanf( speedStr, "%f,%f", speed+0, speed+1 );
                }
                
                int size = 10;
                const char *sizeStr = findKey((const char**)keys, (const char**)values, "size");
                if( sizeStr!=NULL )
                {
                    sscanf( sizeStr, "%d", &size );
                }

                const char *fontStr = findKey((const char**)keys, (const char**)values, "font");
                if( NULL==fontStr )
                {
                    fprintf( stderr, "font is not specified\n" );
                    abort();
                }

                int pos = 0;
                const char *posStr = findKey((const char**)keys, (const char**)values, "pos");
                if( posStr!=NULL )
                {
                    if( ! strcmp(posStr, "left") )
                        pos = 0;
                    else if( ! strcmp(posStr, "center") )
                        pos = 1;
                    else if( ! strcmp(posStr, "right") )
                        pos = 2;
                    else
                    {
                        fprintf( stderr, "pos %s unrecognisze\n", posStr );
                        abort();
                    }
                }

                FT_Library      library;

                if ( FT_Init_FreeType( &library ) )
                {
                    fprintf(stderr, "%s:%d error in FT_Init_FreeType\n", __FILE__, __LINE__);
                    return -1;
                }
                            
                FT_Face face;
                int error = FT_New_Face( library,
                           fontStr, //"/usr/share/fonts/truetype/arial.ttf",
                           0,
                           &face );
                if( error )
                {
                    fprintf(stderr, "cannot load %s\n", fontStr);
                    abort();
                }
                
                ImageWord_t *piw = iwInit(text, size, face, fg, bg);
                FT_Done_Face    ( face );
                
                FT_Done_FreeType( library );
                
                piw->offsetX = offset[0];
                piw->offsetY = offset[1];
                piw->speedX  = speed[0]; 
                piw->speedY  = speed[1]; 
                piw->pos     = pos;
                
                textEffect->piw = piw;
                textEffect->frameNumber = 0;
                
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect  = (void*) textEffect;
                effect->fct     = (void*) imgText;
                effect->free    = (void*) freeText;
                effects[numEffects++] = effect;
            }

            cmd = "scale";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                ScaleEffect_t *scale = bsCalloc( sizeof(ScaleEffect_t) );
                
                const char *scaleStr = findKey((const char**)keys, (const char**)values, "scale");
                if( scaleStr!=NULL )
                {
                    sscanf(scaleStr, "%f,%f", & scale->scaleX, & scale->scaleY);
                }
               
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect  = (void*) scale;
                effect->fct     = (void*) imgScale;
                effects[numEffects++] = effect;
            }
            
            cmd = "bitrate";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                const char *videoStr = findKey((const char**)keys, (const char**)values, "video");
                if( videoStr!=NULL )
                {
                    sscanf( videoStr, "%d", &videoRate );
                }

                const char *audioStr = findKey((const char**)keys, (const char**)values, "audio");
                if( audioStr!=NULL )
                {
                    sscanf( audioStr, "%d", &audioRate );
                }
            }
            
            cmd = "drop";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                DropEffect_t *drop = bsCalloc( sizeof(DropEffect_t) );
                
                const char *value = findKey((const char**)keys, (const char**)values, "drop");
                if( value!=NULL)
                {
                    sscanf( value, "%f", & drop->frequency );
                }

                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) drop;
                effect->fct = (void*) imgDrop;
                effects[numEffects++] = effect;
            }

            cmd = "file";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                const char *tcInStr = findKey((const char**)keys, (const char**)values, "tcIn");
                if( tcInStr!=NULL )
                {
                    tcInFlag = 1;
                    tcStringToTC(&tcIn, tcInStr);
                }
                
                const char *tcOutStr = findKey((const char**)keys, (const char**)values, "tcOut");
                if( tcOutStr!=NULL )
                {
                    tcOutFlag = 1;
                    tcStringToTC(&tcOut, tcOutStr);
                }
               
				const char *inputFileStr = findKey((const char**)keys, (const char**)values, "input");
				if( inputFileStr!=NULL )
					inputFile = bsStrDup( inputFileStr );

				const char *outputFileStr = findKey((const char**)keys, (const char**)values, "output");
				if( outputFileStr!=NULL )
					outputFile = bsStrDup( outputFileStr );

            }

            cmd = "resample";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                ResampleEffect_t *resample = bsCalloc( sizeof(DropEffect_t) );
                
                const char *value = findKey((const char**)keys, (const char**)values, "resample");
                if( value!=NULL)
                {
                    sscanf( value, "%d", & resample->factor );
                }
                
                const char *size = findKey((const char**)keys, (const char**)values, "size");
                if( size!=NULL )
                {
                    sscanf( size, "%d,%d", & resample->width, & resample->height );
                }
                
                Effect_t     *effect = bsCalloc( sizeof(Effect_t) );
                effect->effect = (void*) resample;
                effect->fct = (void*) imgResample;
                effects[numEffects++] = effect;
            }

            cmd = "nosound";
            if( !strncmp(keys[0], cmd, strlen( cmd ) ) )
            {
                noSound = 1;
            }

        }
        
//        if( 0==numEffects )
//            continue;

        if( NULL==inputFile || NULL==outputFile )
        {
            if( 0==numEffects )
                break;
                
            fprintf(stderr, "inputFile or outputFile is not specified\n");
            abort();
        }
        printf("%s -> %s\n", inputFile, outputFile);
        
        //
        ImageSequenceReader_t *pisr = isrOpen(inputFile);
        ImageSequenceWriter_t *pisw = NULL;
        
        if( NULL==pisr )
        {
            fprintf(stderr, "cannot open %s\n", inputFile);
            abort();
        }
        
        int 	numberOfFramesToDo = -1;
        Image_t         *pimg = NULL;
        SoundBuffer_t   *psb = NULL;
        int 	numberOfFrames = -1;

        if( tcInFlag )
            isrSeekToFrame( pisr, (double)tcIn.frameNumber/tcIn.fps);
        if( tcInFlag && tcOutFlag )
            numberOfFrames = numberOfFramesToDo = tcOut.frameNumber-tcIn.frameNumber;
        else if( tcOutFlag )
            numberOfFrames = numberOfFramesToDo = tcOut.frameNumber;
        
//		else
//			numberOfFrames = isrGetNumberOfFrames(pisr);

        int frame;
        for(frame=0; ; frame++)
        {
            if(numberOfFramesToDo!=-1 && frame>=numberOfFramesToDo)
                break;
                
            printf("% 5d / % 5d   \r", frame, numberOfFrames); fflush(stdout);
            if( isrGetFrame( pisr, &pimg, noSound ? NULL : &psb ) < 0 )
                break;
                
            VideoFrame_t    vf1, vf2;
            vf1.pimg    = NULL;
            vf2.pimg    = NULL;
            vf1.psb     = NULL;
            vf2.psb     = NULL;
            int srcFlag = 0;
            
            imgRGBUint8ToARGBFloat32( pimg, & vf1.pimg );
            if( psb!=NULL )
                vf1.psb = sndSBCopy(psb);
			
			vf1.frameNumber = vf2.frameNumber = frame;
			
            int effect;
            for(effect=0; effect<numEffects; effect++)
            {
                if(srcFlag)
                {
                    BSFREEFCT( vf1.pimg, imgFree );
                }
                else
                {
                    BSFREEFCT( vf2.pimg, imgFree );
                }
                
                effects[effect]->fct( srcFlag ? &vf2 : &vf1, srcFlag ? &vf1 : &vf2, effects[effect]->effect );

                if( NULL==(srcFlag ? vf1.psb : vf2.psb) && psb!=NULL )
                {
                    if( srcFlag )
                
                        vf1.psb = sndSBCopy(vf2.psb);
                    else
                        vf2.psb = sndSBCopy(vf1.psb);
                }
                
                srcFlag = !srcFlag;

            }

            Image_t *pimg_Output = NULL;
            imgARGBFloat32ToRGBUint8(srcFlag ? vf2.pimg : vf1.pimg, &pimg_Output);
            
            SoundBuffer_t *psb_Output = srcFlag ? vf2.psb : vf1.psb;
            
            if( NULL==pisw )
            {
                    char    buffer[256];
                    BSDictionary_t *props;
                    
                    props = bsDictNew();

                    sprintf( buffer, "%d,%d", imgGetWidth( pimg_Output ), imgGetHeight( pimg_Output ) );
                    
                    bsDictSetValueForKey( props, bsStrDup(buffer), "video_size",    BSDICTFREE_VALUE );
                    //    bsDictSetValueForKey( props, "mpeg2video",  "video_codec",      BSDICTFREE_NONE );

                    if( videoRate>0 )
                    {
                        sprintf( buffer, "%d", videoRate );
                        bsDictSetValueForKey( props, bsStrDup(buffer),     "video_bitrate",    BSDICTFREE_VALUE );
                    }
                    
                    double fps = isrGetFPS(pisr);
                    if(fps<=0)
                    {
                        fprintf(stderr, "Warning setting fps to 25 by default\n");
                        fps = 25;
                    }
                    sprintf( buffer, "%f", fps );

					bsDictSetValueForKey( props, bsStrDup(buffer),          "video_fps",        BSDICTFREE_NONE );

                    if( psb_Output!= NULL )
                    {
                        sprintf( buffer, "%d", psb->properties.sampleRate );
                        bsDictSetValueForKey( props, bsStrDup(buffer), "audio_rate", BSDICTFREE_VALUE );
                        
                        sprintf( buffer, "%d", psb->properties.numChannels );
                        bsDictSetValueForKey( props, bsStrDup(buffer), "audio_num_channels", BSDICTFREE_VALUE );
                        
                        if( audioRate>0 )
                        {
                            sprintf( buffer, "%d", audioRate );
                            bsDictSetValueForKey( props, bsStrDup(buffer), "audio_bitrate", BSDICTFREE_VALUE );
                        }
                        
                    }
                
                    pisw = iswOpen( outputFile, props );
                    bsDictFree( props );
                    
                    if( NULL==pisw )
                    {
                        fprintf(stderr, "Cannot open output movie %s\n", outputFile);
                        abort();
                    }
            }

            iswPutFrame( pisw, pimg_Output, psb_Output );
            
            BSFREEFCT( pimg_Output,  imgFree );
            BSFREEFCT( vf1.pimg, imgFree );
            BSFREEFCT( vf2.pimg, imgFree );
            BSFREEFCT( vf1.psb, sndSBFree );
            BSFREEFCT( vf2.psb, sndSBFree );
            
        }
        
        BSFREEFCT( pisr, isrClose );
        BSFREEFCT( pisw, iswClose );
        BSFREE( inputFile );
        BSFREE( outputFile );
        
        for(loop=0; loop<numEffects; loop++)
        {
            if( effects[loop]->free!=NULL )
                effects[loop]->free(effects[loop]->effect);
			
			bsFree( effects[loop]->effect );
        }
    }
    

    fclose( file );
    
    return 0;
}
